# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

from dateutil.parser import isoparse


def is_uuid(value: str) -> bool:
    """Check if a string is a valid UUID."""
    try:
        uuid.UUID(value)
        return True
    except (ValueError, TypeError):
        return False


def is_iso_date(value: str) -> bool:
    """Check if a string is a valid ISO 8601 date."""
    try:
        isoparse(value)
        return True
    except (ValueError, TypeError):
        return False


def is_bool_string(value: str) -> bool:
    """Check if a string explicitly represents a boolean value with true/false."""

    return value.strip().lower() in {"true", "false"}



# Copyright 2019-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Annotated, Generator, Optional, TypeVar
from uuid import UUID

from pydantic import BeforeValidator, Field, conlist

from pydantic_forms.validators.components.contact_person import ContactPerson
from pydantic_forms.validators.helpers import remove_empty_items

T = TypeVar("T")  # pragma: no mutate


def customer_contact_list(
    customer_id: Optional[UUID] = None,
    customer_key: Optional[str] = "customer_id",
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
) -> type[list[T]]:
    def json_schema_extra() -> Generator:
        if customer_id:
            yield "customerId", customer_id
        if customer_key:
            yield "customerKey", customer_key

    return Annotated[  # type: ignore
        conlist(ContactPerson, min_length=min_items, max_length=max_items),
        BeforeValidator(remove_empty_items),
        Field(json_schema_extra=dict(json_schema_extra())),
    ]



from more_itertools import first
from strawberry.types.nodes import SelectedField, Selection

from orchestrator.graphql.types import OrchestratorInfo


def get_selected_fields(info: OrchestratorInfo) -> list[str]:
    """Get SelectedField names from the requested query (info).

    Can be used to get the selected fields of the schema, to only fetch those from the database.

    Args:
        info: The info class with request information.

    returns the names of SelectedFields as a list of strings.
    """
    root_selected = info.selected_fields[0]

    def has_field_name(selection: Selection, field_name: str) -> bool:
        return isinstance(selection, SelectedField) and selection.name == field_name

    page_items = first((selection for selection in root_selected.selections if has_field_name(selection, "page")), None)
    if not page_items:
        page_items = root_selected

    return [selection.name for selection in page_items.selections if isinstance(selection, SelectedField)]



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Annotated, Any, Literal

from pydantic import BaseModel, Field, model_validator
from sqlalchemy import DOUBLE_PRECISION, INTEGER, and_
from sqlalchemy import cast as sa_cast
from sqlalchemy.sql.elements import ColumnElement
from typing_extensions import Self

from orchestrator.search.core.types import FilterOp, SQLAColumn


class NumericRange(BaseModel):
    start: int | float
    end: int | float

    @model_validator(mode="after")
    def validate_order(self) -> Self:
        if self.end <= self.start:
            raise ValueError("'end' must be greater than 'start'")
        return self


class NumericValueFilter(BaseModel):
    """A filter for single numeric value comparisons (int or float)."""

    op: Literal[FilterOp.EQ, FilterOp.NEQ, FilterOp.LT, FilterOp.LTE, FilterOp.GT, FilterOp.GTE]
    value: int | float

    def to_expression(self, column: SQLAColumn, path: str) -> ColumnElement[bool]:
        cast_type = INTEGER if isinstance(self.value, int) else DOUBLE_PRECISION
        numeric_column: ColumnElement[Any] = sa_cast(column, cast_type)
        match self.op:

            case FilterOp.EQ:
                return numeric_column == self.value
            case FilterOp.NEQ:
                return numeric_column != self.value
            case FilterOp.LT:
                return numeric_column < self.value
            case FilterOp.LTE:
                return numeric_column <= self.value
            case FilterOp.GT:
                return numeric_column > self.value
            case FilterOp.GTE:
                return numeric_column >= self.value


class NumericRangeFilter(BaseModel):
    """A filter for a range of numeric values (int or float)."""

    op: Literal[FilterOp.BETWEEN]
    value: NumericRange

    def to_expression(self, column: SQLAColumn, path: str) -> ColumnElement[bool]:
        cast_type = INTEGER if isinstance(self.value.start, int) else DOUBLE_PRECISION
        numeric_column: ColumnElement[Any] = sa_cast(column, cast_type)
        return and_(numeric_column >= self.value.start, numeric_column <= self.value.end)


NumericFilter = Annotated[NumericValueFilter | NumericRangeFilter, Field(discriminator="op")]



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Awaitable, Callable
from typing import Any, TypeVar

import structlog
from ag_ui.core import EventType, StateSnapshotEvent
from pydantic_ai import RunContext
from pydantic_ai.ag_ui import StateDeps
from pydantic_ai.exceptions import ModelRetry
from pydantic_ai.messages import ModelRequest, UserPromptPart
from pydantic_ai.toolsets import FunctionToolset

from orchestrator.api.api_v1.endpoints.search import (
    get_definitions,
    list_paths,
    search_processes,
    search_products,
    search_subscriptions,
    search_workflows,
)
from orchestrator.schemas.search import SearchResultsSchema
from orchestrator.search.core.types import ActionType, EntityType, FilterOp
from orchestrator.search.filters import FilterTree
from orchestrator.search.retrieval.exceptions import FilterValidationError, PathNotFoundError
from orchestrator.search.retrieval.validation import validate_filter_tree
from orchestrator.search.schemas.parameters import PARAMETER_REGISTRY, BaseSearchParameters

from .state import SearchState

logger = structlog.get_logger(__name__)


P = TypeVar("P", bound=BaseSearchParameters)

SearchFn = Callable[[P], Awaitable[SearchResultsSchema[Any]]]

SEARCH_FN_MAP: dict[EntityType, SearchFn] = {
    EntityType.SUBSCRIPTION: search_subscriptions,
    EntityType.WORKFLOW: search_workflows,
    EntityType.PRODUCT: search_products,
    EntityType.PROCESS: search_processes,
}

search_toolset: FunctionToolset[StateDeps[SearchState]] = FunctionToolset(max_retries=1)


def last_user_message(ctx: RunContext[StateDeps[SearchState]]) -> str | None:
    for msg in reversed(ctx.messages):
        if isinstance(msg, ModelRequest):
            for part in msg.parts:
                if isinstance(part, UserPromptPart) and isinstance(part.content, str):
                    return part.content
    return None


@search_toolset.tool
async def set_search_parameters(
    ctx: RunContext[StateDeps[SearchState]],
    entity_type: EntityType,
    action: str | ActionType = ActionType.SELECT,
) -> StateSnapshotEvent:
    """Sets the initial search context, like the entity type and the user's query.

    This MUST be the first tool called to start any new search.
    Warning: Calling this tool will erase any existing filters and search results from the state.
    """
    params = ctx.deps.state.parameters or {}
    is_new_search = params.get("entity_type") != entity_type.value
    final_query = (last_user_message(ctx) or "") if is_new_search else params.get("query", "")

    logger.debug(
        "Setting search parameters",
        entity_type=entity_type.value,
        action=action,
        is_new_search=is_new_search,
        query=final_query,
    )

    ctx.deps.state.parameters = {"action": action, "entity_type": entity_type, "filters": None, "query": final_query}
    ctx.deps.state.results = []
    logger.debug("Search parameters set", parameters=ctx.deps.state.parameters)

    return StateSnapshotEvent(
        type=EventType.STATE_SNAPSHOT,
        snapshot=ctx.deps.state.model_dump(),
    )


@search_toolset.tool(retries=2)
async def set_filter_tree(
    ctx: RunContext[StateDeps[SearchState]],
    filters: FilterTree | None,
) -> StateSnapshotEvent:
    """Replace current filters atomically with a full FilterTree, or clear with None.

    Requirements:
    - Root/group operators must be 'AND' or 'OR' (uppercase).
    - Provide either PathFilters or nested groups under `children`.
    - See the FilterTree schema examples for the exact shape.
    """
    if ctx.deps.state.parameters is None:
        raise ModelRetry("Search parameters are not initialized. Call set_search_parameters first.")

    entity_type = EntityType(ctx.deps.state.parameters["entity_type"])

    logger.debug(
        "Setting filter tree",
        entity_type=entity_type.value,
        has_filters=filters is not None,
        filter_summary=f"{len(filters.get_all_leaves())} filters" if filters else "no filters",
    )

    try:
        await validate_filter_tree(filters, entity_type)
    except PathNotFoundError as e:
        logger.debug(f"{PathNotFoundError.__name__}: {str(e)}")
        raise ModelRetry(f"{str(e)} Use discover_filter_paths tool to find valid paths.")
    except FilterValidationError as e:
        # ModelRetry will trigger an agent retry, containing the specific validation error.
        logger.debug(f"Filter validation failed: {str(e)}")
        raise ModelRetry(str(e))
    except Exception as e:
        logger.error("Unexpected Filter validation exception", error=str(e))
        raise ModelRetry(f"Filter validation failed: {str(e)}. Please check your filter structure and try again.")

    filter_data = None if filters is None else filters.model_dump(mode="json", by_alias=True)
    ctx.deps.state.parameters["filters"] = filter_data
    return StateSnapshotEvent(type=EventType.STATE_SNAPSHOT, snapshot=ctx.deps.state.model_dump())


@search_toolset.tool
async def execute_search(
    ctx: RunContext[StateDeps[SearchState]],
    limit: int = 10,
) -> StateSnapshotEvent:
    """Execute the search with the current parameters."""
    if not ctx.deps.state.parameters:
        raise ValueError("No search parameters set")

    entity_type = EntityType(ctx.deps.state.parameters["entity_type"])
    param_class = PARAMETER_REGISTRY.get(entity_type)
    if not param_class:
        raise ValueError(f"Unknown entity type: {entity_type}")

    params = param_class(**ctx.deps.state.parameters)
    logger.debug(
        "Executing database search",
        search_entity_type=entity_type.value,
        limit=limit,
        has_filters=params.filters is not None,
        query=params.query,
        action=params.action,
    )

    if params.filters:
        logger.debug("Search filters", filters=params.filters)

    params.limit = limit

    fn = SEARCH_FN_MAP[entity_type]
    search_results = await fn(params)

    logger.debug(
        "Search completed",
        total_results=len(search_results.data) if search_results.data else 0,
    )

    ctx.deps.state.results = search_results.data

    return StateSnapshotEvent(type=EventType.STATE_SNAPSHOT, snapshot=ctx.deps.state.model_dump())


@search_toolset.tool
async def discover_filter_paths(
    ctx: RunContext[StateDeps[SearchState]],
    field_names: list[str],
    entity_type: EntityType | None = None,
) -> dict[str, dict[str, Any]]:
    """Discovers available filter paths for a list of field names.

    Returns a dictionary where each key is a field_name from the input list and
    the value is its discovery result.
    """
    if not entity_type and ctx.deps.state.parameters:
        entity_type = EntityType(ctx.deps.state.parameters.get("entity_type"))
    if not entity_type:
        entity_type = EntityType.SUBSCRIPTION

    all_results = {}
    for field_name in field_names:
        paths_response = await list_paths(prefix="", q=field_name, entity_type=entity_type, limit=100)

        matching_leaves = []
        for leaf in paths_response.leaves:
            if field_name.lower() in leaf.name.lower():
                matching_leaves.append(
                    {
                        "name": leaf.name,
                        "value_kind": leaf.ui_types,
                        "paths": leaf.paths,
                    }
                )

        matching_components = []
        for comp in paths_response.components:
            if field_name.lower() in comp.name.lower():
                matching_components.append(
                    {
                        "name": comp.name,
                        "value_kind": comp.ui_types,
                    }
                )

        result_for_field: dict[str, Any]
        if not matching_leaves and not matching_components:
            result_for_field = {
                "status": "NOT_FOUND",
                "guidance": f"No filterable paths found containing '{field_name}'. Do not create a filter for this.",
                "leaves": [],
                "components": [],
            }
        else:
            result_for_field = {
                "status": "OK",
                "guidance": f"Found {len(matching_leaves)} field(s) and {len(matching_components)} component(s) for '{field_name}'.",
                "leaves": matching_leaves,
                "components": matching_components,
            }

        all_results[field_name] = result_for_field
    logger.debug("Returning found fieldname - path mapping", all_results=all_results)
    return all_results


@search_toolset.tool
async def get_valid_operators() -> dict[str, list[FilterOp]]:
    """Gets the mapping of field types to their valid filter operators."""
    definitions = await get_definitions()

    operator_map = {}
    for ui_type, type_def in definitions.items():
        key = ui_type.value

        if hasattr(type_def, "operators"):
            operator_map[key] = type_def.operators
    return operator_map



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Sequence

import structlog
from sqlalchemy.engine.row import RowMapping
from sqlalchemy.orm import Session

from orchestrator.search.core.types import FilterOp, SearchMetadata
from orchestrator.search.filters import FilterTree, LtreeFilter
from orchestrator.search.schemas.parameters import BaseSearchParameters
from orchestrator.search.schemas.results import MatchingField, SearchResponse, SearchResult

from .builder import build_candidate_query
from .pagination import PaginationParams
from .retrievers import Retriever
from .utils import generate_highlight_indices

logger = structlog.get_logger(__name__)


def _format_response(
    db_rows: Sequence[RowMapping], search_params: BaseSearchParameters, metadata: SearchMetadata
) -> SearchResponse:
    """Format database query results into a `SearchResponse`.

    Converts raw SQLAlchemy `RowMapping` objects into `SearchResult` instances,
    including highlight metadata if present in the database results.

    Args:
        db_rows (Sequence[RowMapping]): The rows returned from the executed SQLAlchemy query.
        search_params (BaseSearchParameters): The search parameters, including query text and filters.
        metadata (SearchMetadata): Metadata about the search execution.

    Returns:
        SearchResponse: A list of `SearchResult` objects containing entity IDs, scores,
        and optional highlight information.
    """

    if not db_rows:
        return SearchResponse(results=[], metadata=metadata)

    user_query = search_params.query

    results = []
    for row in db_rows:
        matching_field = None

        if (
            user_query
            and (text := row.get(Retriever.HIGHLIGHT_TEXT_LABEL))
            and (path := row.get(Retriever.HIGHLIGHT_PATH_LABEL))
        ):
            if not isinstance(text, str):
                text = str(text)
            if not isinstance(path, str):
                path = str(path)

            highlight_indices = generate_highlight_indices(text, user_query) or None
            matching_field = MatchingField(text=text, path=path, highlight_indices=highlight_indices)

        elif not user_query and search_params.filters and metadata.search_type == "structured":
            # Structured search (filter-only)
            matching_field = _extract_matching_field_from_filters(search_params.filters)

        results.append(
            SearchResult(
                entity_id=str(row.entity_id),
                score=row.score,
                perfect_match=row.get("perfect_match", 0),
                matching_field=matching_field,
            )
        )
    return SearchResponse(results=results, metadata=metadata)


def _extract_matching_field_from_filters(filters: FilterTree) -> MatchingField | None:
    """Extract the first path filter to use as matching field for structured searches.

    TODO: Should we allow a list of matched fields in the MatchingField model?
    We need a different approach, probably a cross join in StructuredRetriever.
    """
    leaves = filters.get_all_leaves()
    if len(leaves) != 1:
        return None

    pf = leaves[0]

    if isinstance(pf.condition, LtreeFilter):
        op = pf.condition.op
        # Prefer the original component/pattern (validator may set path="*" and move the value)
        display = str(getattr(pf.condition, "value", "") or pf.path)

        # There can be no match for abscence.
        if op == FilterOp.NOT_HAS_COMPONENT:
            return None

        return MatchingField(text=display, path=display, highlight_indices=[(0, len(display))])

    # Everything thats not Ltree
    val = getattr(pf.condition, "value", "")
    text = "" if val is None else str(val)
    return MatchingField(text=text, path=pf.path, highlight_indices=[(0, len(text))])


async def execute_search(
    search_params: BaseSearchParameters,
    db_session: Session,
    pagination_params: PaginationParams | None = None,
) -> SearchResponse:
    """Execute a hybrid search and return ranked results.

    Builds a candidate entity query based on the given search parameters,
    applies the appropriate ranking strategy, and executes the final ranked
    query to retrieve results.

    Args:
        search_params (BaseSearchParameters): The search parameters specifying vector, fuzzy, or filter criteria.
        db_session (Session): The active SQLAlchemy session for executing the query.
        pagination_params (PaginationParams): Parameters controlling pagination of the search results.
        limit (int, optional): The maximum number of search results to return, by default 5.

    Returns:
        SearchResponse: A list of `SearchResult` objects containing entity IDs, scores,
        and optional highlight metadata.

    Notes:
        If no vector query, filters, or fuzzy term are provided, a warning is logged
        and an empty result set is returned.
    """

    if not search_params.vector_query and not search_params.filters and not search_params.fuzzy_term:
        logger.warning("No search criteria provided (vector_query, fuzzy_term, or filters).")
        return SearchResponse(results=[], metadata=SearchMetadata.empty())

    candidate_query = build_candidate_query(search_params)

    pagination_params = pagination_params or PaginationParams()
    retriever = await Retriever.from_params(search_params, pagination_params)
    logger.debug("Using retriever", retriever_type=retriever.__class__.__name__)

    final_stmt = retriever.apply(candidate_query)
    final_stmt = final_stmt.limit(search_params.limit)
    logger.debug(final_stmt)
    result = db_session.execute(final_stmt).mappings().all()

    return _format_response(result, search_params, retriever.metadata)



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Literal, overload

from fastapi import APIRouter, HTTPException, Query, status
from sqlalchemy import case, select
from sqlalchemy.orm import selectinload

from orchestrator.db import (
    ProcessTable,
    ProductTable,
    WorkflowTable,
    db,
)
from orchestrator.domain.base import SubscriptionModel
from orchestrator.schemas.search import (
    PageInfoSchema,
    PathsResponse,
    ProcessSearchResult,
    ProcessSearchSchema,
    ProductSearchResult,
    ProductSearchSchema,
    SearchResultsSchema,
    SubscriptionSearchResult,
    WorkflowSearchResult,
    WorkflowSearchSchema,
)
from orchestrator.search.core.exceptions import InvalidCursorError
from orchestrator.search.core.types import EntityType, UIType
from orchestrator.search.filters.definitions import generate_definitions
from orchestrator.search.indexing.registry import ENTITY_CONFIG_REGISTRY
from orchestrator.search.retrieval import execute_search
from orchestrator.search.retrieval.builder import build_paths_query, create_path_autocomplete_lquery, process_path_rows
from orchestrator.search.retrieval.pagination import (
    create_next_page_cursor,
    process_pagination_cursor,
)
from orchestrator.search.retrieval.validation import is_lquery_syntactically_valid
from orchestrator.search.schemas.parameters import (
    BaseSearchParameters,
    ProcessSearchParameters,
    ProductSearchParameters,
    SubscriptionSearchParameters,
    WorkflowSearchParameters,
)
from orchestrator.search.schemas.results import SearchResult, TypeDefinition
from orchestrator.services.subscriptions import format_special_types

router = APIRouter()


def _create_search_result_item(
    entity: WorkflowTable | ProductTable | ProcessTable, entity_type: EntityType, search_info: SearchResult
) -> WorkflowSearchResult | ProductSearchResult | ProcessSearchResult | None:
    match entity_type:
        case EntityType.WORKFLOW:
            workflow_data = WorkflowSearchSchema.model_validate(entity)
            return WorkflowSearchResult(
                workflow=workflow_data,
                score=search_info.score,
                perfect_match=search_info.perfect_match,
                matching_field=search_info.matching_field,
            )
        case EntityType.PRODUCT:
            product_data = ProductSearchSchema.model_validate(entity)
            return ProductSearchResult(
                product=product_data,
                score=search_info.score,
                perfect_match=search_info.perfect_match,
                matching_field=search_info.matching_field,
            )
        case EntityType.PROCESS:
            process_data = ProcessSearchSchema.model_validate(entity)
            return ProcessSearchResult(
                process=process_data,
                score=search_info.score,
                perfect_match=search_info.perfect_match,
                matching_field=search_info.matching_field,
            )
        case _:
            return None


@overload
async def _perform_search_and_fetch(
    search_params: BaseSearchParameters,
    entity_type: Literal[EntityType.WORKFLOW],
    eager_loads: list[Any],
    cursor: str | None = None,
) -> SearchResultsSchema[WorkflowSearchResult]: ...


@overload
async def _perform_search_and_fetch(
    search_params: BaseSearchParameters,
    entity_type: Literal[EntityType.PRODUCT],
    eager_loads: list[Any],
    cursor: str | None = None,
) -> SearchResultsSchema[ProductSearchResult]: ...


@overload
async def _perform_search_and_fetch(
    search_params: BaseSearchParameters,
    entity_type: Literal[EntityType.PROCESS],
    eager_loads: list[Any],
    cursor: str | None = None,
) -> SearchResultsSchema[ProcessSearchResult]: ...


async def _perform_search_and_fetch(
    search_params: BaseSearchParameters,
    entity_type: EntityType,
    eager_loads: list[Any],
    cursor: str | None = None,
) -> SearchResultsSchema[Any]:
    try:
        pagination_params = await process_pagination_cursor(cursor, search_params)
    except InvalidCursorError:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Invalid pagination cursor")

    search_response = await execute_search(
        search_params=search_params,
        db_session=db.session,
        pagination_params=pagination_params,
    )
    if not search_response.results:
        return SearchResultsSchema(search_metadata=search_response.metadata)

    next_page_cursor = create_next_page_cursor(search_response.results, pagination_params, search_params.limit)
    has_next_page = next_page_cursor is not None
    page_info = PageInfoSchema(has_next_page=has_next_page, next_page_cursor=next_page_cursor)

    config = ENTITY_CONFIG_REGISTRY[entity_type]
    entity_ids = [res.entity_id for res in search_response.results]
    pk_column = getattr(config.table, config.pk_name)
    ordering_case = case({entity_id: i for i, entity_id in enumerate(entity_ids)}, value=pk_column)

    stmt = select(config.table).options(*eager_loads).filter(pk_column.in_(entity_ids)).order_by(ordering_case)
    entities = db.session.scalars(stmt).all()

    search_info_map = {res.entity_id: res for res in search_response.results}
    data = []
    for entity in entities:
        entity_id = getattr(entity, config.pk_name)
        search_info = search_info_map.get(str(entity_id))
        if not search_info:
            continue

        search_result_item = _create_search_result_item(entity, entity_type, search_info)
        if search_result_item:
            data.append(search_result_item)

    return SearchResultsSchema(data=data, page_info=page_info, search_metadata=search_response.metadata)


@router.post(
    "/subscriptions",
    response_model=SearchResultsSchema[SubscriptionSearchResult],
)
async def search_subscriptions(
    search_params: SubscriptionSearchParameters,
    cursor: str | None = None,
) -> SearchResultsSchema[SubscriptionSearchResult]:
    try:
        pagination_params = await process_pagination_cursor(cursor, search_params)
    except InvalidCursorError:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Invalid pagination cursor")

    search_response = await execute_search(
        search_params=search_params,
        db_session=db.session,
        pagination_params=pagination_params,
    )

    if not search_response.results:
        return SearchResultsSchema(search_metadata=search_response.metadata)

    next_page_cursor = create_next_page_cursor(search_response.results, pagination_params, search_params.limit)
    has_next_page = next_page_cursor is not None
    page_info = PageInfoSchema(has_next_page=has_next_page, next_page_cursor=next_page_cursor)

    search_info_map = {res.entity_id: res for res in search_response.results}
    results_data = []
    for sub_id, search_info in search_info_map.items():
        subscription_model = SubscriptionModel.from_subscription(sub_id)
        sub_data = subscription_model.model_dump(exclude_unset=False)
        search_result_item = SubscriptionSearchResult(
            subscription=format_special_types(sub_data),
            score=search_info.score,
            perfect_match=search_info.perfect_match,
            matching_field=search_info.matching_field,
        )
        results_data.append(search_result_item)

    return SearchResultsSchema(data=results_data, page_info=page_info, search_metadata=search_response.metadata)


@router.post("/workflows", response_model=SearchResultsSchema[WorkflowSearchResult])
async def search_workflows(
    search_params: WorkflowSearchParameters,
    cursor: str | None = None,
) -> SearchResultsSchema[WorkflowSearchResult]:
    return await _perform_search_and_fetch(
        search_params=search_params,
        entity_type=EntityType.WORKFLOW,
        eager_loads=[selectinload(WorkflowTable.products)],
        cursor=cursor,
    )


@router.post("/products", response_model=SearchResultsSchema[ProductSearchResult])
async def search_products(
    search_params: ProductSearchParameters,
    cursor: str | None = None,
) -> SearchResultsSchema[ProductSearchResult]:
    return await _perform_search_and_fetch(
        search_params=search_params,
        entity_type=EntityType.PRODUCT,
        eager_loads=[
            selectinload(ProductTable.workflows),
            selectinload(ProductTable.fixed_inputs),
            selectinload(ProductTable.product_blocks),
        ],
        cursor=cursor,
    )


@router.post("/processes", response_model=SearchResultsSchema[ProcessSearchResult])
async def search_processes(
    search_params: ProcessSearchParameters,
    cursor: str | None = None,
) -> SearchResultsSchema[ProcessSearchResult]:
    return await _perform_search_and_fetch(
        search_params=search_params,
        entity_type=EntityType.PROCESS,
        eager_loads=[
            selectinload(ProcessTable.workflow),
        ],
        cursor=cursor,
    )


@router.get(
    "/paths",
    response_model=PathsResponse,
    response_model_exclude_none=True,
)
async def list_paths(
    prefix: str = Query("", min_length=0),
    q: str | None = Query(None, description="Query for path suggestions"),
    entity_type: EntityType = Query(EntityType.SUBSCRIPTION),
    limit: int = Query(10, ge=1, le=10),
) -> PathsResponse:

    if prefix:
        lquery_pattern = create_path_autocomplete_lquery(prefix)

        if not is_lquery_syntactically_valid(lquery_pattern, db.session):
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail=f"Prefix '{prefix}' creates an invalid search pattern.",
            )
    stmt = build_paths_query(entity_type=entity_type, prefix=prefix, q=q)
    stmt = stmt.limit(limit)
    rows = db.session.execute(stmt).all()

    leaves, components = process_path_rows(rows)
    return PathsResponse(leaves=leaves, components=components)


@router.get(
    "/definitions",
    response_model=dict[UIType, TypeDefinition],
    response_model_exclude_none=True,
)
async def get_definitions() -> dict[UIType, TypeDefinition]:
    """Provide a static definition of operators and schemas for each UI type."""
    return generate_definitions()



#!/usr/bin/env python3
"""The main application module.

This module contains the main `AgenticOrchestratorCore` class for the `FastAPI` backend and
provides the ability to run the CLI.
"""
# Copyright 2019-2025 SURF
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import typer
from pydantic_ai.models.openai import OpenAIModel
from pydantic_ai.toolsets import FunctionToolset
from structlog import get_logger

from orchestrator.app import OrchestratorCore
from orchestrator.cli.main import app as cli_app
from orchestrator.llm_settings import LLMSettings, llm_settings

logger = get_logger(__name__)


class AgenticOrchestratorCore(OrchestratorCore):
    def __init__(
        self,
        *args: Any,
        llm_model: OpenAIModel | str = "gpt-4o-mini",
        llm_settings: LLMSettings = llm_settings,
        agent_tools: list[FunctionToolset] | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the `AgenticOrchestratorCore` class.

        This class takes the same arguments as the `OrchestratorCore` class.

        Args:
            *args: All the normal arguments passed to the `OrchestratorCore` class.
            llm_model: An OpenAI model class or string, not limited to OpenAI models (gpt-4o-mini etc)
            llm_settings: A class of settings for the LLM
            agent_tools: A list of tools that can be used by the agent
            **kwargs: Additional arguments passed to the `OrchestratorCore` class.

        Returns:
            None
        """
        self.llm_model = llm_model
        self.agent_tools = agent_tools
        self.llm_settings = llm_settings

        super().__init__(*args, **kwargs)

        logger.info("Mounting the agent")
        self.register_llm_integration()

    def register_llm_integration(self) -> None:
        """Mount the Agent endpoint.

        This helper mounts the agent endpoint on the application.

        Returns:
            None

        """
        from orchestrator.search.agent import build_agent_app

        agent_app = build_agent_app(self.llm_model, self.agent_tools)
        self.mount("/agent", agent_app)


main_typer_app = typer.Typer()
main_typer_app.add_typer(cli_app, name="orchestrator", help="The orchestrator CLI commands")

if __name__ == "__main__":
    main_typer_app()



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import requests
import structlog
from requests.exceptions import RequestException
from sqlalchemy import select, text
from sqlalchemy.exc import SQLAlchemyError

from orchestrator.db import EngineSettingsTable, db
from orchestrator.schemas.engine_settings import EngineSettingsSchema, GlobalStatusEnum
from orchestrator.settings import app_settings

logger = structlog.get_logger(__name__)


def get_engine_settings() -> EngineSettingsTable:
    """Returns the EngineSettingsTable object. Raises an exception if the query does not return exactly one row."""
    return db.session.execute(select(EngineSettingsTable)).scalar_one()


def get_engine_settings_for_update() -> EngineSettingsTable:
    """Same as get_engine_settings but blocks until transactions on engine_settings table are committed."""
    return db.session.execute(select(EngineSettingsTable).with_for_update()).scalar_one()


def generate_engine_global_status(engine_settings: EngineSettingsTable) -> GlobalStatusEnum:
    """Returns the global status of the engine."""
    if engine_settings.global_lock and engine_settings.running_processes > 0:
        return GlobalStatusEnum.PAUSING
    if engine_settings.global_lock and engine_settings.running_processes == 0:
        return GlobalStatusEnum.PAUSED
    return GlobalStatusEnum.RUNNING


def post_update_to_slack(engine_status: EngineSettingsSchema, user: str) -> None:
    """Post engine settings update to slack.

    Args:
        engine_status: EngineStatus
        user: The user who executed the change

    Returns:
        None

    """
    try:
        if engine_status.global_lock is True:
            action = f"stopped the `{app_settings.ENVIRONMENT}` workflow engine. The orchestrator will pause all running processes."
        else:
            action = f"started the `{app_settings.ENVIRONMENT}` workflow engine. The orchestrator will pick up all pending processes."

        message = {"text": f"User `{user}` {action}"}
        requests.post(app_settings.SLACK_ENGINE_SETTINGS_HOOK_URL, json=message, timeout=5)

    # Catch all Request exceptions and log. Then pass
    except RequestException:
        logger.exception("Post to slack failed.")
        pass


def reset_search_index(*, tx_commit: bool = False) -> None:
    try:
        db.session.execute(text("REFRESH MATERIALIZED VIEW CONCURRENTLY subscriptions_search;"))
    except SQLAlchemyError as e:
        logger.error("Something went wrong while refreshing materialized view", msg=str(e))
        raise e
    finally:
        if tx_commit:
            db.session.commit()
    return



"""Helper script to rewrite your orchestrator codebase for orchestrator-core 2.0.0.

Refer to the 2.0 migration guide documentation for background.
"""

import re
import sys
from pathlib import Path
from typing import Iterable

from orchestrator.devtools.scripts.shared import (
    find_and_remove_aliases,
    has_word,
    insert_import,
    migrate,
    move_import,
    remove_imports,
    replace_words,
)


def rewrite_subscription_instance_lists(f: Path) -> list[str]:
    """Rewrite all SubscriptionInstanceList occurrences in a file and fix the imports.

    Refer to the "SubscriptionInstanceList" section in the 2.0 migration guide for details.

    Returns:
        names of any replaced SI Lists.
    """
    text = text_orig = f.read_text()
    sil = "SubscriptionInstanceList"
    if not has_word(text, sil):
        return []

    # Remove aliases in imports and in usages
    text, aliases = find_and_remove_aliases(text, sil)
    text = replace_words(text, aliases, sil) if aliases else text

    # Remove imports
    text, _changed = remove_imports(text, "orchestrator.domain.base", sil)

    # Replace SubscriptionInstanceList classes with types
    rgx = (
        r"^\s*class (?P<name>\w+)\([^)]*SubscriptionInstanceList\[(?P<type>(?:\w+|Union\[[\s\w,]+\]))\][^)]*\):\n"
        r"(?:^\s+min_items = (?P<min_items>(?:\w+|\d+))\n)*"
        r"(?:^\s+max_items = (?P<max_items>(?:\w+|\d+))\n)*"
    )

    def replace_si_list(match: re.Match) -> str:
        subscript_type = match.group("type")
        result = "\n%s = Annotated[list[%s], Len(%s)]\n"

        def len_params() -> Iterable[str]:
            if min_items := match.group("min_items"):
                yield f"min_length={min_items}"
            if max_items := match.group("max_items"):
                yield f"max_length={max_items}"

        return result % (match.group("name"), subscript_type, ", ".join(len_params()))

    names = [match[0] for match in re.findall(rgx, text, flags=re.M)]
    text = re.sub(rgx, replace_si_list, text, flags=re.M)

    symbols = {
        "Len": "from annotated_types import Len",
        "Annotated": "from typing import Annotated",
    }

    for symbol, import_stmt in symbols.items():
        if has_word(text_orig, symbol):
            # Assumes that if the symbol is already in the file, it was also imported
            continue
        if not has_word(text, symbol):
            # Not every symbol may be necessary
            continue
        text = insert_import(text, import_stmt)

    with f.open(mode="w"):
        f.write_text(text)
    return names


re_serializable_property = re.compile(r"^(\s+)(@serializable_property)([^\n]*)\n", flags=re.MULTILINE)


def replace_serializable_props(f: Path) -> bool:
    """Replace @serializable_property with pydantic's @computed_field and updates imports.

    As an example, this changes:

      @serializable_property
      def title():
        ...

    To:

      @computed_field  # type: ignore[misc]
      @property
      def title():
        ...

    The type:ignore is recommended by pydantic to silence mypy; if you don't use mypy it's not needed.
    """
    text = f.read_text()
    text, changed = remove_imports(text, "orchestrator.domain.base", "serializable_property")
    if not changed:
        return False
    text = insert_import(text, "from pydantic import computed_field")
    text = re_serializable_property.sub(r"\1@computed_field  # type: ignore[misc]\3\1@property\n", text)
    with f.open(mode="w"):
        f.write_text(text)
    return True


def migrate_file(f: Path) -> bool:
    imports = {
        "SI": move_import(f, "SI", "orchestrator.domain.base", "orchestrator.types"),
        "VlanRanges": move_import(f, "VlanRanges", "orchestrator.utils.vlans", "nwastdlib.vlans"),
        "ReadOnlyField_forms": move_import(f, "ReadOnlyField", "pydantic_forms.core", "pydantic_forms.validators"),
        "ReadOnlyField_core": move_import(f, "ReadOnlyField", "orchestrator.forms", "pydantic_forms.validators"),
        "pydantic BaseSettings": move_import(f, "BaseSettings", "pydantic", "pydantic_settings"),
    }
    lines = []
    if replaced_lists := ", ".join(rewrite_subscription_instance_lists(f)):
        lines.append(f"replaced subscription instance lists [{replaced_lists}]")
    if replace_serializable_props(f):
        lines.append("replaced serializable properties")
    lines.extend([f"Moved {k} import" for k, v in imports.items() if v])

    if lines:
        formatted_lines = "\n".join(f" - {line}" for line in lines)
        print(f"Updated {f.name:50s}\n{formatted_lines}")

    return bool(lines)


if __name__ == "__main__":
    try:
        _target_dir = Path(sys.argv[1])
        assert _target_dir.is_dir()
    except Exception:
        print("Need a directory as parameter")
        sys.exit(1)

    sys.exit(0 if migrate(_target_dir, migrate_file) else 1)



# Copyright 2019-2025 SURF, ESnet, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
from collections import defaultdict
from datetime import datetime
from inspect import get_annotations, isclass
from itertools import groupby, zip_longest
from operator import attrgetter
from typing import (
    Any,
    Callable,
    ClassVar,
    Iterable,
    Mapping,
    Optional,
    TypeVar,
    Union,
    cast,
    get_args,
    get_type_hints,
)
from uuid import UUID, uuid4

import structlog
from more_itertools import bucket, first, flatten, one, only
from pydantic import BaseModel, ConfigDict, Field, ValidationError
from pydantic.fields import PrivateAttr
from sqlalchemy import select
from sqlalchemy.orm import joinedload, selectinload

from orchestrator.db import (
    ProductBlockTable,
    ProductTable,
    SubscriptionInstanceRelationTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    SubscriptionTable,
    db,
)
from orchestrator.db.queries.subscription_instance import get_subscription_instance_dict
from orchestrator.domain.helpers import (
    _to_product_block_field_type_iterable,
    get_root_blocks_to_instance_ids,
    no_private_attrs,
)
from orchestrator.domain.lifecycle import (
    ProductLifecycle,
    lookup_specialized_type,
    register_specialized_type,
    validate_lifecycle_status,
)
from orchestrator.domain.subscription_instance_transform import field_transformation_rules, transform_instance_fields
from orchestrator.services.products import get_product_by_id
from orchestrator.types import (
    SAFE_USED_BY_TRANSITIONS_FOR_STATUS,
    SubscriptionLifecycle,
    filter_nonetype,
    get_origin_and_args,
    get_possible_product_block_types,
    is_list_type,
    is_of_type,
    is_optional_type,
    is_union_type,
    list_factory,
)
from orchestrator.utils.datetime import nowtz
from orchestrator.utils.docs import make_product_block_docstring, make_subscription_model_docstring
from pydantic_forms.types import State, UUIDstr

logger = structlog.get_logger(__name__)


class ProductNotInRegistryError(Exception):
    pass


T = TypeVar("T")  # pragma: no mutate
S = TypeVar("S", bound="SubscriptionModel")  # pragma: no mutate
B = TypeVar("B", bound="ProductBlockModel")  # pragma: no mutate


class DomainModel(BaseModel):
    """Base class for domain models.

    Contains all common Product block/Subscription instance code
    """

    model_config = ConfigDict(validate_assignment=True, validate_default=True)

    __base_type__: ClassVar[type["DomainModel"] | None] = None  # pragma: no mutate
    _product_block_fields_: ClassVar[dict[str, Any]]
    _non_product_block_fields_: ClassVar[dict[str, type]]

    def __init_subclass__(cls, *args: Any, lifecycle: list[SubscriptionLifecycle] | None = None, **kwargs: Any) -> None:
        pass

    def __eq__(self, other: Any) -> bool:
        # PrivateAttr fields are excluded from both objects during the equality check.
        # Added for #652 primarily because ProductBlockModel._db_model is now lazy loaded.
        with no_private_attrs(self), no_private_attrs(other):
            return super().__eq__(other)

    @classmethod
    def __pydantic_init_subclass__(
        cls,
        *args: Any,
        lifecycle: list[SubscriptionLifecycle] | None = None,
        **kwargs: Any,
    ) -> None:
        super().__pydantic_init_subclass__()
        cls._find_special_fields()

        if kwargs.keys():
            logger.warning(
                "Unexpected keyword arguments in domain model class",  # pragma: no mutate
                class_name=cls.__name__,
                kwargs=kwargs.keys(),
            )

        if not lifecycle:
            return

        # Check if dependency subscription instance models conform to the same lifecycle
        for product_block_field_name, product_block_field_type in cls._get_depends_on_product_block_types().items():
            field_types = _to_product_block_field_type_iterable(product_block_field_type)

            for lifecycle_status, field_type in itertools.product(lifecycle, field_types):
                validate_lifecycle_status(product_block_field_name, field_type, lifecycle_status)

    @classmethod
    def _get_depends_on_product_block_types(
        cls,
    ) -> dict[str, type["ProductBlockModel"] | tuple[type["ProductBlockModel"]]]:
        """Return all the product block model types.

        This strips any List[], Optional[] or Annotated[] types.
        """
        result = {}
        for product_block_field_name, product_block_field_type in cls._product_block_fields_.items():
            field_type: type["ProductBlockModel"] | tuple[type["ProductBlockModel"]]
            if is_union_type(product_block_field_type) and not is_optional_type(product_block_field_type):
                # exclude non-Optional Unions as they contain more than one useful element.
                _origin, args = get_origin_and_args(product_block_field_type)
                field_type = cast(tuple[type[ProductBlockModel]], args)
            elif is_list_type(product_block_field_type) or (
                is_optional_type(product_block_field_type) and len(get_args(product_block_field_type)) <= 2
            ):
                _origin, args = get_origin_and_args(product_block_field_type)
                field_type = first(args)
            else:
                field_type = product_block_field_type

            result[product_block_field_name] = field_type
        return result

    @classmethod
    def _find_special_fields(cls) -> None:
        """Make and store a list of resource_type fields and product block fields."""
        cls._non_product_block_fields_ = {}
        cls._product_block_fields_ = {}

        annotations = get_annotations(cls)

        # Retrieve type hints with evaluated ForwardRefs (for nested blocks) and extra annotations
        type_hints = get_type_hints(cls, localns={cls.__name__: cls}, include_extras=True)

        # But this also returns inherited fields so cross-check against the annotations
        final_annotations = {k: type_hints[k] for k in annotations}

        for field_name, field_type in final_annotations.items():
            if field_name.startswith("_"):
                continue

            try:
                is_product_block_field = (
                    is_union_type(field_type, DomainModel)
                    or is_list_type(field_type, DomainModel)
                    or is_optional_type(field_type, DomainModel)
                    or is_of_type(field_type, DomainModel)
                )
            except TypeError:
                # issubclass does not work on typing types
                is_product_block_field = False

            # Figure out if this field_name has an alias. Needed sometimes for serializable properties that
            # have a 'real' property with the same name that has a field alias.
            if (field := cls.model_fields.get(field_name)) and field.alias:
                field_name = field.alias

            # We only want fields that are on this class and not on the related product blocks
            if is_product_block_field:
                cls._product_block_fields_[field_name] = field_type
            else:
                cls._non_product_block_fields_[field_name] = field_type

    @classmethod
    def _init_instances(  # noqa: C901
        cls, subscription_id: UUID, skip_keys: set[str] | None = None
    ) -> dict[str, Union[list["ProductBlockModel"], "ProductBlockModel", None]]:
        """Initialize default subscription instances.

        When a new domain model is created that is not loaded from an existing subscription.
        We also create all subscription instances for it. This function does that.

        Args:
            subscription_id: The UUID of the subscription
            skip_keys: set of fields on the class to skip when creating dummy instances.

        Returns:
            A dict with instances to pass to the new model
        """
        if skip_keys is None:
            skip_keys = set()

        product_block_field_names = set(cls._product_block_fields_) - skip_keys
        return {
            product_block_field_name: cls._init_instance(product_block_field_name, subscription_id)
            for product_block_field_name in product_block_field_names
        }

    @classmethod
    def _init_instance(
        cls, product_block_field_name: str, subscription_id: UUID
    ) -> Union["ProductBlockModel", list["ProductBlockModel"], None]:
        """Initialize a default subscription instance."""
        product_block_field_type = cls._product_block_fields_[product_block_field_name]

        if is_list_type(product_block_field_type):
            return list_factory(product_block_field_type, subscription_id=subscription_id)

        if is_optional_type(product_block_field_type, ProductBlockModel):
            return None

        if is_union_type(product_block_field_type):
            raise ValueError(
                "Union Types must always be `Optional` when calling `.new().` We are unable to detect which "
                "type to intialise and Union types always cross subscription boundaries."
            )

        product_block_model = product_block_field_type
        # Scalar field of a ProductBlockModel expects 1 instance
        return product_block_model.new(subscription_id=subscription_id)

    @classmethod
    def _load_instances(  # noqa: C901
        cls,
        db_instances: list[SubscriptionInstanceTable],
        status: SubscriptionLifecycle,
        match_domain_attr: bool = True,
        in_use_by_id_boundary: UUID | None = None,
    ) -> dict[str, Optional["ProductBlockModel"] | list["ProductBlockModel"]]:
        """Load subscription instances for this domain model.

        When a new domain model is loaded from an existing subscription we also load all
        subscription instances for it. This function does that.

        Args:
            db_instances: list of database models to load from
            status: SubscriptionLifecycle of subscription to check if models match
            match_domain_attr: Match domain attribute on relations [1]
            in_use_by_id_boundary: Match domain attribute on relations with this in_use_by_id [1]

        Note [1]: only use these parameters when loading product blocks that are in use by another product block.

        Returns:
            A dict with instances to pass to the new model

        """

        instances: dict[str, ProductBlockModel | None | list[ProductBlockModel]] = {}

        def keyfunc(i: SubscriptionInstanceTable) -> str:
            return i.product_block.name

        sorted_instances = sorted(db_instances, key=keyfunc)
        grouped_instances = {k: list(g) for k, g in groupby(sorted_instances, keyfunc)}

        def match_domain_model_attr_if_possible(field_name: str) -> Callable:
            def domain_filter(instance: SubscriptionInstanceTable) -> bool:
                """Match domain model attributes.

                This helper is necessary to filter through all relations in a subscription. Not all subscriptions have a
                domain model attribute that is set as it is not always necessary. However when it is set, it is necessary
                to filter through instances depending on that attribute.

                Args:
                    instance: depends on subscription instance

                Returns:
                    Boolean of match.

                """
                # We don't match on the product_blocks directly under subscriptions. They don't have in_use_by relations to those
                if not match_domain_attr:
                    return True

                def include_relation(relation: SubscriptionInstanceRelationTable) -> bool:
                    return bool(relation.domain_model_attr) and (
                        not in_use_by_id_boundary or relation.in_use_by_id == in_use_by_id_boundary
                    )

                attr_names = {
                    rel.domain_model_attr for rel in instance.in_use_by_block_relations if include_relation(rel)
                }

                # We can assume true if no domain_model_attr is set.
                return not attr_names or field_name in attr_names

            return domain_filter

        product_block_model_list: list[ProductBlockModel] = []
        for product_block_field_name, product_block_field_type in cls._product_block_fields_.items():
            filter_func = match_domain_model_attr_if_possible(product_block_field_name)

            possible_product_block_types = flatten_product_block_types(product_block_field_type)
            field_type_names = list(possible_product_block_types.keys())
            filtered_instances = flatten([grouped_instances.get(name, []) for name in field_type_names])
            instance_list = list(filter(filter_func, filtered_instances))

            if is_list_type(product_block_field_type):
                if product_block_field_name not in grouped_instances:
                    product_block_model_list = []

                product_block_model_list.extend(
                    possible_product_block_types[instance.product_block.name].from_db(
                        subscription_instance=instance, status=status
                    )
                    for instance in instance_list
                )

                instances[product_block_field_name] = product_block_model_list
            else:
                instance = only(instance_list)
                if not is_optional_type(product_block_field_type) and instance is None:
                    raise ValueError("Required subscription instance is missing in database")

                if is_optional_type(product_block_field_type) and instance is None:
                    instances[product_block_field_name] = None
                elif instance:
                    assert (  # noqa: S101
                        len(possible_product_block_types) is not None
                    ), "Product block model has not been resolved. Unable to continue"
                    instances[product_block_field_name] = possible_product_block_types[
                        instance.product_block.name
                    ].from_db(subscription_instance=instance, status=status)
        return instances

    @classmethod
    def _data_from_lifecycle(cls, other: "DomainModel", status: SubscriptionLifecycle, subscription_id: UUID) -> dict:
        data = other.model_dump()

        for field_name, field_type in cls._product_block_fields_.items():
            value = getattr(other, field_name)
            if value is None:
                continue

            _origin, args = get_origin_and_args(field_type)
            if is_list_type(field_type):
                data[field_name] = []
                list_field_type = one(args)
                possible_product_block_types = get_possible_product_block_types(list_field_type)

                for item in value:
                    data[field_name].append(
                        possible_product_block_types[item.name]._from_other_lifecycle(item, status, subscription_id)
                    )
            else:
                if is_union_type(field_type):
                    if is_optional_type(field_type):
                        data[field_name] = None
                    for f_type in filter_nonetype(args):
                        if f_type.name == value.name:
                            field_type = f_type
                            break
                    else:
                        logger.warning(
                            "Cannot determine type for product block field value",
                            field_name=field_name,
                            field_type=field_type,
                            value_name=value.name,
                        )
                        continue

                data[field_name] = field_type._from_other_lifecycle(value, status, subscription_id)
        return data

    def _save_instances(
        self, subscription_id: UUID, status: SubscriptionLifecycle
    ) -> tuple[list[SubscriptionInstanceTable], dict[str, list[SubscriptionInstanceTable]]]:
        """Save subscription instances for this domain model.

        When a domain model is saved to the database we need to save all depends_on subscription instances for it.

        Args:
            subscription_id: The subscription id
            status: SubscriptionLifecycle of subscription to check if models match

        Returns:
            A list with instances which are saved and a dict with direct depends_on relations.

        """
        saved_instances: list[SubscriptionInstanceTable] = []
        depends_on_instances: dict[str, list[SubscriptionInstanceTable]] = {}

        self._check_duplicate_instance_relations()

        for product_block_field, product_block_field_type in self._product_block_fields_.items():
            product_block_models = getattr(self, product_block_field)
            if is_list_type(product_block_field_type):
                field_instance_list = []
                for product_block_model in product_block_models:
                    saved, depends_on_instance = product_block_model.save(
                        subscription_id=subscription_id, status=status
                    )
                    field_instance_list.append(depends_on_instance)
                    saved_instances.extend(saved)
                depends_on_instances[product_block_field] = field_instance_list
            elif (
                is_optional_type(product_block_field_type) or is_union_type(product_block_field_type)
            ) and product_block_models is None:
                pass
            else:
                saved, depends_on_instance = product_block_models.save(subscription_id=subscription_id, status=status)
                depends_on_instances[product_block_field] = [depends_on_instance]
                saved_instances.extend(saved)

        return saved_instances, depends_on_instances

    def _check_duplicate_instance_relations(self) -> None:
        """Check that there are no product block fields referring to the same instance.

        A ValueError is raised if this is the case.
        """

        def get_id(product_block: ProductBlockModel) -> UUID:
            return product_block.subscription_instance_id

        def get_ids(field_name: str) -> Iterable[tuple[str, UUID]]:
            match getattr(self, field_name):
                case list() as value_list:
                    blocks = (value for value in value_list if isinstance(value, ProductBlockModel))
                    yield from ((f"{field_name}.{index}", get_id(block)) for index, block in enumerate(blocks))
                case ProductBlockModel() as block:
                    yield field_name, get_id(block)

        def to_fields(mm: Iterable[tuple[str, UUID]]) -> list[str]:
            return [x[0] for x in mm]

        field_id_tuples = flatten(get_ids(field_name) for field_name in self._product_block_fields_)
        id_buckets = bucket(field_id_tuples, lambda x: x[1])
        id_fields_tuples = ((id_, to_fields(id_buckets[id_])) for id_ in id_buckets)
        duplicates = [(id_, fields) for id_, fields in id_fields_tuples if len(fields) > 1]
        if duplicates:
            details = "; ".join(f"instance {id_} is used in fields {fields}" for id_, fields in duplicates)
            raise ValueError(f"Cannot link the same subscription instance multiple times: {details}")


def flatten_product_block_types(product_block_field_type: Any) -> dict[str, type["ProductBlockModel"]]:
    """Extract product block types and return mapping of product block names to product block classes."""
    product_block_model: Any = product_block_field_type
    if is_list_type(product_block_field_type):
        _origin, args = get_origin_and_args(product_block_field_type)
        product_block_model = one(args)
    return get_possible_product_block_types(product_block_model)


def get_depends_on_product_block_type_list(
    product_block_types: dict[str, type["ProductBlockModel"] | tuple[type["ProductBlockModel"]]],
) -> list[type["ProductBlockModel"]]:
    product_blocks_types_in_model = []
    for product_block_type in product_block_types.values():
        if is_union_type(product_block_type):
            _origin, args = get_origin_and_args(product_block_type)
            product_blocks_types_in_model.extend(list(filter_nonetype(args)))
        else:
            product_blocks_types_in_model.append(product_block_type)

    if product_blocks_types_in_model and isinstance(first(product_blocks_types_in_model), tuple):
        return one(product_blocks_types_in_model)

    return product_blocks_types_in_model


class ProductBlockModel(DomainModel):
    r"""This is the base class for all product block models.

    This class should have been called SubscriptionInstanceModel.

    ProductTable Blocks are represented as dataclasses with pydantic runtime validation.

    Different stages of a subscription lifecycle could require different product block
    definition.Mainly to support mandatory fields when a subscription is active. To support
    this a lifecycle specific product block definition can be created by subclassing the
    generic product block with keyword argument 'lifecycle' and overriding its fields.

    All product blocks are related to a database ProductBlockTable object through the `product_block_name`
    that is given as class keyword argument.

    Define a product block:

        >>> class BlockInactive(ProductBlockModel, product_block_name="Virtual Circuit"):
        ...    int_field: Optional[int] = None
        ...    str_field: Optional[str] = None

        >>> class Block(BlockInactive, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        ...    int_field: int
        ...    str_field: str

    This example defines a product_block with two different contraints based on lifecycle. `Block` is valid only for `ACTIVE`
    And `BlockInactive` for all other states.
    `product_block_name` must be defined on the base class and need not to be defined on the others

    Create a new empty product block:

        >>> example1 = BlockInactive()  # doctest: +SKIP

    Create a new instance based on a dict in the state:

        >>> example2 = BlockInactive(**state)  # doctest:+SKIP

    To retrieve a ProductBlockModel from the database:

        >>> BlockInactive.from_db(subscription_instance_id)  # doctest:+SKIP
    """

    registry: ClassVar[dict[str, type["ProductBlockModel"]]] = {}  # pragma: no mutate
    __names__: ClassVar[set[str]] = set()
    product_block_id: ClassVar[UUID]
    description: ClassVar[str]
    tag: ClassVar[str]
    _db_model: SubscriptionInstanceTable | None = PrivateAttr(default=None)

    # Product block name. This needs to be an instance var because its part of the API (we expose it to the frontend)
    # Is actually optional since abstract classes don't have it.
    # TODO #427 name is used as both a ClassVar and a pydantic Field, for which Pydantic 2.x raises
    #  warnings (which may become errors)
    name: str | None
    subscription_instance_id: UUID
    owner_subscription_id: UUID
    label: str | None = None

    @classmethod
    def _fix_pb_data(cls) -> None:
        if not cls.name:
            raise ValueError(f"Cannot create instance of abstract class. Use one of {cls.__names__}")

        # Would have been nice to do this in __init_subclass__ but that runs outside the app context so we can't
        # access the db. So now we do it just before we instantiate the instance
        if not hasattr(cls, "product_block_id"):
            product_block = db.session.scalars(
                select(ProductBlockTable).filter(ProductBlockTable.name == cls.name)
            ).one()
            cls.product_block_id = product_block.product_block_id
            cls.description = product_block.description
            cls.tag = product_block.tag

    @classmethod
    def __pydantic_init_subclass__(  # type: ignore[override]
        cls,
        *,
        product_block_name: str | None = None,
        lifecycle: list[SubscriptionLifecycle] | None = None,
        **kwargs: Any,
    ) -> None:
        super().__pydantic_init_subclass__(lifecycle=lifecycle, **kwargs)

        if product_block_name is not None:
            # This is a concrete product block base class (so not a abstract super class or a specific lifecycle version)
            cls.name = product_block_name
            cls.__base_type__ = cls
            cls.__names__ = {cls.name}
            cls.registry[cls.name] = cls
        elif lifecycle is None:
            # Abstract class, no product block name
            cls.name = None
            cls.__names__ = set()

        # For everything except abstract classes
        if cls.name is not None:
            register_specialized_type(cls, lifecycle)

            # Add ourselves to any super class. That way we can match a superclass to an instance when loading
            for klass in cls.__mro__:
                if issubclass(klass, ProductBlockModel):
                    klass.__names__.add(cls.name)

        cls.__doc__ = make_product_block_docstring(cls, lifecycle)

    @classmethod
    def diff_product_block_in_database(cls) -> dict[str, set[str]]:
        """Return any differences between the attrs defined on the domain model and those on product blocks in the database.

        This is only needed to check if the domain model and database models match which would be done during testing...
        """
        if not cls.name:
            # This is a superclass we can't check that
            return {}

        product_block_db = db.session.scalars(
            select(ProductBlockTable).where(ProductBlockTable.name == cls.name)
        ).one_or_none()
        product_blocks_in_db = {pb.name for pb in product_block_db.depends_on} if product_block_db else set()

        product_blocks_in_model = cls._get_depends_on_product_block_types()
        product_blocks_types_in_model = get_depends_on_product_block_type_list(product_blocks_in_model)

        product_blocks_in_model = set(
            flatten(map(attrgetter("__names__"), product_blocks_types_in_model))
        )  # type: ignore

        missing_product_blocks_in_db = product_blocks_in_model - product_blocks_in_db  # type: ignore
        missing_product_blocks_in_model = product_blocks_in_db - product_blocks_in_model  # type: ignore

        resource_types_model = set(cls._non_product_block_fields_)
        resource_types_db = {rt.resource_type for rt in product_block_db.resource_types} if product_block_db else set()

        missing_resource_types_in_db = resource_types_model - resource_types_db
        missing_resource_types_in_model = resource_types_db - resource_types_model

        logger.debug(
            "ProductBlockTable blocks diff",
            product_block_db=product_block_db.name if product_block_db else None,
            product_blocks_in_db=product_blocks_in_db,
            product_blocks_in_model=product_blocks_in_model,
            resource_types_db=resource_types_db,
            resource_types_model=resource_types_model,
            missing_product_blocks_in_db=missing_product_blocks_in_db,
            missing_product_blocks_in_model=missing_product_blocks_in_model,
            missing_resource_types_in_db=missing_resource_types_in_db,
            missing_resource_types_in_model=missing_resource_types_in_model,
        )

        missing_data: dict[str, Any] = {}
        for product_block_model in product_blocks_types_in_model:
            if product_block_model.name == cls.name or product_block_model.name in missing_data:
                continue
            missing_data.update(product_block_model.diff_product_block_in_database())

        diff: dict[str, set[str]] = {
            k: v
            for k, v in {
                "missing_product_blocks_in_db": missing_product_blocks_in_db,
                "missing_product_blocks_in_model": missing_product_blocks_in_model,
                "missing_resource_types_in_db": missing_resource_types_in_db,
                "missing_resource_types_in_model": missing_resource_types_in_model,
            }.items()
            if v
        }

        if diff:
            missing_data[cls.name] = diff

        return missing_data

    @classmethod
    def new(cls: type[B], subscription_id: UUID, **kwargs: Any) -> B:
        """Create a new empty product block.

        We need to use this instead of the normal constructor because that assumes you pass in
        all required values. That is cumbersome since that means creating a tree of product blocks.

        This is similar to `from_product_id()`
        """
        sub_instances = cls._init_instances(subscription_id, set(kwargs.keys()))

        subscription_instance_id = uuid4()

        # Make sure product block stuff is already set if new is the first usage of this class
        cls._fix_pb_data()

        db_model = SubscriptionInstanceTable(
            product_block_id=cls.product_block_id,
            subscription_instance_id=subscription_instance_id,
            subscription_id=subscription_id,
        )
        db.session.enable_relationship_loading(db_model)

        if kwargs_name := kwargs.pop("name", None):
            # Not allowed to change the product block model name at runtime. This is only possible through
            # the `product_block_name=..` metaclass parameter
            logger.warning("Ignoring `name` keyword to ProductBlockModel.new()", rejected_name=kwargs_name)
        model = cls(
            name=cls.name,
            subscription_instance_id=subscription_instance_id,
            owner_subscription_id=subscription_id,
            label=db_model.label,
            **sub_instances,
            **kwargs,
        )
        model.db_model = db_model
        return model

    @classmethod
    def _load_instances_values(cls, instance_values: list[SubscriptionInstanceValueTable]) -> dict[str, str]:
        """Load non product block fields (instance values).

        Args:
            instance_values: List of instance values from database

        Returns:
            Dict of fields to use for constructor

        """
        instance_values_dict: State = {}
        list_field_names = set()

        # Set default values
        for field_name, field_type in cls._non_product_block_fields_.items():
            # Ensure that empty lists are handled OK
            if is_list_type(field_type):
                instance_values_dict[field_name] = []
                list_field_names.add(field_name)
            elif is_optional_type(field_type):
                # Initialize "optional required" fields
                instance_values_dict[field_name] = None

        for siv in instance_values:
            # check the type of the siv in the instance and act accordingly: only lists and scalar values supported
            resource_type_name = siv.resource_type.resource_type
            if resource_type_name in list_field_names:
                instance_values_dict[resource_type_name].append(siv.value)
            else:
                instance_values_dict[resource_type_name] = siv.value

        return instance_values_dict

    @classmethod
    def _from_other_lifecycle(
        cls: type[B],
        other: "ProductBlockModel",
        status: SubscriptionLifecycle,
        subscription_id: UUID,
    ) -> B:
        """Create new domain model from instance while changing the status.

        This makes sure we always have a specific instance.
        """
        if not cls.__base_type__:
            cls = ProductBlockModel.registry.get(other.name, cls)  # type:ignore
            cls = lookup_specialized_type(cls, status)

        data = cls._data_from_lifecycle(other, status, subscription_id)

        cls._fix_pb_data()
        model = cls(**data)
        model.db_model = other.db_model
        return model

    @classmethod
    def from_db(
        cls: type[B],
        subscription_instance_id: UUID | None = None,
        subscription_instance: SubscriptionInstanceTable | None = None,
        status: SubscriptionLifecycle | None = None,
    ) -> B:
        """Create a product block based on a subscription instance from the database.

        This function is similar to `from_subscription()`

            >>> subscription_instance_id = KNOWN_UUID_IN_DB  # doctest:+SKIP
            >>> si_from_db = db.SubscriptionInstanceTable.query.get(subscription_instance_id)  # doctest:+SKIP
            >>> example3 = ProductBlockModel.from_db(subscription_instance=si_from_db)  # doctest:+SKIP
            >>> example4 = ProductBlockModel.from_db(subscription_instance_id=subscription_instance_id)  # doctest:+SKIP
        """
        # Fill values from actual subscription
        if subscription_instance_id:
            subscription_instance = db.session.get(SubscriptionInstanceTable, subscription_instance_id)
        if subscription_instance:
            subscription_instance_id = subscription_instance.subscription_instance_id
        assert subscription_instance_id  # noqa: S101
        assert subscription_instance  # noqa: S101

        if not status:
            status = SubscriptionLifecycle(subscription_instance.subscription.status)

        if not cls.__base_type__:
            cls = ProductBlockModel.registry.get(subscription_instance.product_block.name, cls)  # type:ignore
            cls = lookup_specialized_type(cls, status)

        elif not issubclass(cls, lookup_specialized_type(cls, status)):
            raise ValueError(f"{cls} is not valid for lifecycle {status}")

        label = subscription_instance.label

        instance_values = cls._load_instances_values(subscription_instance.values)
        sub_instances = cls._load_instances(
            subscription_instance.depends_on,
            status,
            match_domain_attr=True,
            in_use_by_id_boundary=subscription_instance_id,
        )

        cls._fix_pb_data()
        try:
            model = cls(
                name=cls.name,
                subscription_instance_id=subscription_instance_id,
                owner_subscription_id=subscription_instance.subscription_id,
                label=label,
                subscription=subscription_instance.subscription,
                **instance_values,  # type: ignore
                **sub_instances,
            )
            model.db_model = subscription_instance

            return model
        except ValidationError:
            logger.exception(
                "Subscription is not correct in database",
                loaded_instance_values=instance_values,
                loaded_sub_instances=sub_instances,
            )
            raise

    def _save_instance_values(
        self, product_block: ProductBlockTable, current_values: list[SubscriptionInstanceValueTable]
    ) -> list[SubscriptionInstanceValueTable]:
        """Save non product block fields (instance values).

        Returns:
            List of database instances values to save

        """
        resource_types = {rt.resource_type: rt for rt in product_block.resource_types}
        current_values_dict: dict[str, list[SubscriptionInstanceValueTable]] = defaultdict(list)
        for siv in current_values:
            current_values_dict[siv.resource_type.resource_type].append(siv)

        subscription_instance_values = []
        for field_name, field_type in self._non_product_block_fields_.items():
            assert (  # noqa: S101
                field_name in resource_types
            ), f"Domain model {self.__class__} does not match the ProductBlockTable {product_block.name}, missing: {field_name} {resource_types}"

            resource_type = resource_types[field_name]
            value = getattr(self, field_name)
            if value is None:
                continue
            if is_list_type(field_type):
                for val, siv in zip_longest(value, current_values_dict[field_name]):
                    if val is not None:
                        if siv:
                            siv.value = str(val)
                            subscription_instance_values.append(siv)
                        else:
                            subscription_instance_values.append(
                                SubscriptionInstanceValueTable(resource_type=resource_type, value=str(val))
                            )
            else:
                if field_name in current_values_dict:
                    current_value = current_values_dict[field_name][0]
                    current_value.value = str(value)
                    subscription_instance_values.append(current_value)
                else:
                    subscription_instance_values.append(
                        SubscriptionInstanceValueTable(resource_type=resource_type, value=str(value))
                    )
        return subscription_instance_values

    def _set_instance_domain_model_attrs(
        self,
        subscription_instance: SubscriptionInstanceTable,
        subscription_instance_mapping: dict[str, list[SubscriptionInstanceTable]],
    ) -> None:
        """Save the domain model attribute to the database.

        This function iterates through the subscription instances and stores the domain model attribute in the
        hierarchy relationship.

        Args:
            subscription_instance: The subscription instance object.
            subscription_instance_mapping: a mapping of the domain model attribute a underlying instances

        Returns:
            None

        """
        depends_on_block_relations = []
        # Set the domain_model_attrs in the database
        for domain_model_attr, instances in subscription_instance_mapping.items():
            instance: SubscriptionInstanceTable
            for index, instance in enumerate(instances):
                relation = SubscriptionInstanceRelationTable(
                    in_use_by_id=subscription_instance.subscription_instance_id,
                    depends_on_id=instance.subscription_instance_id,
                    order_id=index,
                    domain_model_attr=domain_model_attr,
                )
                depends_on_block_relations.append(relation)
        subscription_instance.depends_on_block_relations = depends_on_block_relations

    def save(
        self, *, subscription_id: UUID, status: SubscriptionLifecycle
    ) -> tuple[list[SubscriptionInstanceTable], SubscriptionInstanceTable]:
        """Save the current model instance to the database.

        This means saving the whole tree of subscription instances and separately saving all instance
        values for this instance. This is called automatically when you return a subscription to the state
        in a workflow step.

        Args:
            status: current SubscriptionLifecycle to check if all constraints match
            subscription_id: Optional subscription id needed if this is a new model

        Returns:
            List of saved instances

        """
        if not self.name:
            raise ValueError(f"Cannot create instance of abstract class. Use one of {self.__names__}")

        # Make sure we have a valid subscription instance database model
        subscription_instance: SubscriptionInstanceTable | None = db.session.get(
            SubscriptionInstanceTable, self.subscription_instance_id
        )
        if subscription_instance:
            # Make sure we do not use a mapped session.
            db.session.refresh(subscription_instance)

            # If this is a "foreign" instance we just stop saving and return it so only its relation is saved
            # We should not touch these themselves
            if self.owner_subscription_id != subscription_id:
                return [], subscription_instance

            self.db_model = subscription_instance
        elif subscription_instance := self.db_model:
            # We only need to add to the session if the subscription_instance does not exist.
            db.session.add(subscription_instance)
        else:
            raise ValueError("Cannot save ProductBlockModel without a db_model")

        subscription_instance.subscription_id = subscription_id

        db.session.flush()

        # Everything is ok, make sure we are of the right class
        specialized_type = lookup_specialized_type(self.__class__, status)
        if specialized_type and not isinstance(self, specialized_type):
            raise ValueError(
                f"Lifecycle status {status} requires specialized type {specialized_type!r}, was: {type(self)!r}"
            )

        # Actually save stuff
        subscription_instance.label = self.label
        subscription_instance.values = self._save_instance_values(
            subscription_instance.product_block, subscription_instance.values
        )

        sub_instances, depends_on_instances = self._save_instances(subscription_id, status)

        # Save the subscription instances relations.
        self._set_instance_domain_model_attrs(subscription_instance, depends_on_instances)

        return sub_instances + [subscription_instance], subscription_instance

    @property
    def subscription(self) -> SubscriptionTable | None:
        return self.db_model.subscription if self.db_model else None

    @property
    def db_model(self) -> SubscriptionInstanceTable | None:
        if not self._db_model:
            self._db_model = db.session.execute(
                select(SubscriptionInstanceTable).where(
                    SubscriptionInstanceTable.subscription_instance_id == self.subscription_instance_id
                )
            ).scalar_one_or_none()
        return self._db_model

    @db_model.setter
    def db_model(self, value: SubscriptionInstanceTable) -> None:
        self._db_model = value

    @property
    def in_use_by(self) -> list[SubscriptionInstanceTable]:  # TODO check where used, might need eagerloading
        """This provides a list of product blocks that depend on this product block."""
        return self.db_model.in_use_by if self.db_model else []

    @property
    def depends_on(self) -> list[SubscriptionInstanceTable]:  # TODO check where used, might need eagerloading
        """This provides a list of product blocks that this product block depends on."""
        return self.db_model.depends_on if self.db_model else []


class ProductModel(BaseModel):
    """Represent the product as defined in the database as a dataclass."""

    model_config = ConfigDict(validate_assignment=True, validate_default=True)

    product_id: UUID
    name: str
    description: str
    product_type: str
    tag: str
    status: ProductLifecycle
    created_at: datetime | None = None
    end_date: datetime | None = None


class SubscriptionModel(DomainModel):
    r"""This is the base class for all product subscription models.

    To use this class, see the examples below:

    Definining a subscription model:

        >>> class SubscriptionInactive(SubscriptionModel, product_type="SP"):  # doctest:+SKIP
        ...    block: Optional[ProductBlockModelInactive] = None

        >>> class Subscription(BlockInactive, lifecycle=[SubscriptionLifecycle.ACTIVE]):  # doctest:+SKIP
        ...    block: ProductBlockModel

    This example defines a subscription model with two different contraints based on lifecycle. `Subscription` is valid only for `ACTIVE`
    And `SubscriptionInactive` for all other states.
    `product_type` must be defined on the base class and need not to be defined on the others

    Create a new empty subscription:

        >>> example1 = SubscriptionInactive.from_product_id(product_id, customer_id)  # doctest:+SKIP

    Create a new instance based on a dict in the state:

        >>> example2 = SubscriptionInactive(**state)  # doctest:+SKIP

    To retrieve a ProductBlockModel from the database:

        >>> SubscriptionInactive.from_subscription(subscription_id)  # doctest:+SKIP
    """

    __model_dump_cache__: ClassVar[dict[UUID, "SubscriptionModel"] | None] = None

    product: ProductModel
    customer_id: str
    _db_model: SubscriptionTable | None = PrivateAttr(default=None)
    subscription_id: UUID = Field(default_factory=uuid4)  # pragma: no mutate
    description: str = "Initial subscription"  # pragma: no mutate
    status: SubscriptionLifecycle = SubscriptionLifecycle.INITIAL  # pragma: no mutate
    insync: bool = False  # pragma: no mutate
    start_date: datetime | None = None  # pragma: no mutate
    end_date: datetime | None = None  # pragma: no mutate
    note: str | None = None  # pragma: no mutate
    version: int = 1  # pragma: no mutate

    def __new__(cls, *args: Any, status: SubscriptionLifecycle | None = None, **kwargs: Any) -> "SubscriptionModel":
        # status can be none if created during change_lifecycle
        if status and not issubclass(cls, lookup_specialized_type(cls, status)):
            raise ValueError(f"{cls} is not valid for status {status}")

        return super().__new__(cls)

    @classmethod
    def __pydantic_init_subclass__(  # type: ignore[override]
        cls, is_base: bool = False, lifecycle: list[SubscriptionLifecycle] | None = None, **kwargs: Any
    ) -> None:
        super().__pydantic_init_subclass__(lifecycle=lifecycle, **kwargs)

        if is_base:
            cls.__base_type__ = cls

        if is_base or lifecycle:
            register_specialized_type(cls, lifecycle)

        cls.__doc__ = make_subscription_model_docstring(cls, lifecycle)

    @classmethod
    def diff_product_in_database(cls, product_id: UUID) -> dict[str, dict[str, set[str] | dict[str, set[str]]]]:
        """Return any differences between the attrs defined on the domain model and those on product blocks in the database.

        This is only needed to check if the domain model and database models match which would be done during testing...
        """
        product_db = db.session.get(ProductTable, product_id)
        product_blocks_in_db = {pb.name for pb in product_db.product_blocks} if product_db else set()

        product_blocks_in_model = cls._get_depends_on_product_block_types()
        product_blocks_types_in_model = get_depends_on_product_block_type_list(product_blocks_in_model)

        product_blocks_in_model = set(
            flatten(map(attrgetter("__names__"), product_blocks_types_in_model))
        )  # type: ignore

        missing_product_blocks_in_db = product_blocks_in_model - product_blocks_in_db  # type: ignore
        missing_product_blocks_in_model = product_blocks_in_db - product_blocks_in_model  # type: ignore

        fixed_inputs_model = set(cls._non_product_block_fields_)
        fixed_inputs_in_db = {fi.name for fi in product_db.fixed_inputs} if product_db else set()

        missing_fixed_inputs_in_db = fixed_inputs_model - fixed_inputs_in_db
        missing_fixed_inputs_in_model = fixed_inputs_in_db - fixed_inputs_model

        logger.debug(
            "ProductTable blocks diff",
            product_block_db=product_db.name if product_db else None,
            product_blocks_in_db=product_blocks_in_db,
            product_blocks_in_model=product_blocks_in_model,
            fixed_inputs_in_db=fixed_inputs_in_db,
            fixed_inputs_model=fixed_inputs_model,
            missing_product_blocks_in_db=missing_product_blocks_in_db,
            missing_product_blocks_in_model=missing_product_blocks_in_model,
            missing_fixed_inputs_in_db=missing_fixed_inputs_in_db,
            missing_fixed_inputs_in_model=missing_fixed_inputs_in_model,
        )

        missing_data_depends_on_blocks: dict[str, set[str]] = {}
        for product_block_in_model in product_blocks_types_in_model:
            missing_data_depends_on_blocks.update(product_block_in_model.diff_product_block_in_database())

        diff: dict[str, set[str] | dict[str, set[str]]] = {
            k: v
            for k, v in {
                "missing_product_blocks_in_db": missing_product_blocks_in_db,
                "missing_product_blocks_in_model": missing_product_blocks_in_model,
                "missing_fixed_inputs_in_db": missing_fixed_inputs_in_db,
                "missing_fixed_inputs_in_model": missing_fixed_inputs_in_model,
                "missing_in_depends_on_blocks": missing_data_depends_on_blocks,
            }.items()
            if v
        }

        missing_data: dict[str, dict[str, set[str] | dict[str, set[str]]]] = {}
        if diff and product_db:
            missing_data[product_db.name] = diff

        return missing_data

    @classmethod
    def _load_root_instances(
        cls,
        subscription_id: UUID | UUIDstr,
    ) -> dict[str, Optional[dict] | list[dict]]:
        """Load root subscription instance(s) for this subscription model.

        When a new subscription model is loaded from an existing subscription, this function loads the entire root
        subscription instance(s) from database using an optimized postgres function. The result of that function
        is used to instantiate the root product block(s).

        The "old" method DomainModel._load_instances() would recursively load subscription instances from the
        database and individually instantiate nested blocks, more or less "manually" reconstructing the subscription.

        The "new" method SubscriptionModel._load_root_instances() takes a different approach; since it has all
        data for the root subscription instance, it can rely on Pydantic to instantiate the root block and all
        nested blocks in one go. This is also why it does not have the params `status` and `match_domain_attr` because
        this information is already encoded in the domain model of a product.
        """
        root_block_instance_ids = get_root_blocks_to_instance_ids(subscription_id)

        root_block_types = {
            field_name: list(flatten_product_block_types(product_block_type).keys())
            for field_name, product_block_type in cls._product_block_fields_.items()
        }

        def get_instances_by_block_names(block_names: list[str]) -> Iterable[dict]:
            for block_name in block_names:
                for instance_id in root_block_instance_ids.get(block_name, []):
                    yield get_subscription_instance_dict(instance_id)

        # Map root product block fields to subscription instance(s) dicts
        instances = {
            field_name: list(get_instances_by_block_names(block_names))
            for field_name, block_names in root_block_types.items()
        }

        # Transform values according to domain models (list[dict] -> dict, add None as default for optionals)
        rules = {
            klass.name: field_transformation_rules(klass) for klass in ProductBlockModel.registry.values() if klass.name
        }
        for instance_list in instances.values():
            for instance in instance_list:
                transform_instance_fields(rules, instance)

        # Support the (theoretical?) usecase of a list of root product blocks
        def unpack_instance_list(field_name: str, instance_list: list[dict]) -> list[dict] | dict | None:
            field_type = cls._product_block_fields_[field_name]
            if is_list_type(field_type):
                return instance_list
            return only(instance_list)

        return {
            field_name: unpack_instance_list(field_name, instance_list)
            for field_name, instance_list in instances.items()
        }

    @classmethod
    def from_product_id(
        cls: type[S],
        product_id: UUID | UUIDstr,
        customer_id: str,
        status: SubscriptionLifecycle = SubscriptionLifecycle.INITIAL,
        description: str | None = None,
        insync: bool = False,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
        note: str | None = None,
        version: int = 1,
    ) -> S:
        """Use product_id (and customer_id) to return required fields of a new empty subscription."""
        # Caller wants a new instance and provided a product_id and customer_id
        product_db = db.session.get(ProductTable, product_id)
        if not product_db:
            raise KeyError("Could not find a product for the given product_id")

        product = ProductModel(
            product_id=product_db.product_id,
            name=product_db.name,
            description=product_db.description,
            product_type=product_db.product_type,
            tag=product_db.tag,
            status=product_db.status,
            created_at=product_db.created_at,
            end_date=product_db.end_date,
        )

        if description is None:
            description = f"Initial subscription of {product.description}"

        subscription_id = uuid4()
        subscription = SubscriptionTable(
            subscription_id=subscription_id,
            product_id=product_id,
            customer_id=customer_id,
            description=description,
            status=status.value,
            insync=insync,
            start_date=start_date,
            end_date=end_date,
            note=note,
            version=version,
        )
        db.session.add(subscription)

        fixed_inputs = {fi.name: fi.value for fi in product_db.fixed_inputs}
        instances = cls._init_instances(subscription_id)

        model = cls(
            product=product,
            customer_id=customer_id,
            subscription_id=subscription_id,
            description=description,
            status=status,
            insync=insync,
            start_date=start_date,
            end_date=end_date,
            note=note,
            version=version,
            **fixed_inputs,
            **instances,
        )
        model.db_model = subscription
        return model

    @classmethod
    def from_other_lifecycle(
        cls: type[S],
        other: "SubscriptionModel",
        status: SubscriptionLifecycle,
        skip_validation: bool = False,
    ) -> S:
        """Create new domain model from instance while changing the status.

        This makes sure we always have a specific instance.
        """
        if not cls.__base_type__:
            # Import here to prevent cyclic imports
            from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY

            cls = SUBSCRIPTION_MODEL_REGISTRY.get(other.product.name, cls)  # type:ignore
            cls = lookup_specialized_type(cls, status)

        # this will raise ValueError when wrong lifecycle transitions are detected in the new domain model
        if not skip_validation:
            validate_lifecycle_change(other, status)

        data = cls._data_from_lifecycle(other, status, other.subscription_id)
        data["status"] = status
        if data["start_date"] is None and status == SubscriptionLifecycle.ACTIVE:
            data["start_date"] = nowtz()
        if data["end_date"] is None and status == SubscriptionLifecycle.TERMINATED:
            data["end_date"] = nowtz()

        model = cls(**data)
        model.db_model = other._db_model

        return model

    # Some common functions shared by from_other_product and from_subscription
    @classmethod
    def _get_subscription(cls: type[S], subscription_id: UUID | UUIDstr) -> SubscriptionTable | None:

        if not isinstance(subscription_id, UUID | UUIDstr):
            raise TypeError(f"subscription_id is of type {type(subscription_id)} instead of UUID | UUIDstr")

        loaders = [
            joinedload(SubscriptionTable.product).selectinload(ProductTable.fixed_inputs),
        ]

        return db.session.get(SubscriptionTable, subscription_id, options=loaders)

    @classmethod
    def _to_product_model(cls: type[S], product: ProductTable) -> ProductModel:
        return ProductModel(
            product_id=product.product_id,
            name=product.name,
            description=product.description,
            product_type=product.product_type,
            tag=product.tag,
            status=product.status,
            created_at=product.created_at if product.created_at else None,
            end_date=product.end_date if product.end_date else None,
        )

    @classmethod
    def from_other_product(
        cls: type[S],
        old_instantiation: S,
        new_product_id: UUID | str,
        new_root: tuple[str, ProductBlockModel] | None = None,
    ) -> S:
        db_product = get_product_by_id(new_product_id)
        if not db_product:
            raise KeyError("Could not find a product for the given product_id")

        old_subscription_id = old_instantiation.subscription_id
        if not (subscription := cls._get_subscription(old_subscription_id)):
            raise ValueError(f"Subscription with id: {old_subscription_id}, does not exist")
        product = cls._to_product_model(db_product)

        status = SubscriptionLifecycle(subscription.status)

        if not cls.__base_type__:
            # Import here to prevent cyclic imports
            from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY

            cls = SUBSCRIPTION_MODEL_REGISTRY.get(subscription.product.name, cls)  # type:ignore
            cls = lookup_specialized_type(cls, status)
        elif not issubclass(cls, lookup_specialized_type(cls, status)):
            raise ValueError(f"{cls} is not valid for lifecycle {status}")

        fixed_inputs = {fi.name: fi.value for fi in db_product.fixed_inputs}

        if new_root:
            name, product_block = new_root
            instances = {name: product_block}
        else:
            # TODO test using cls._load_root_instances() here as well
            instances = cls._load_instances(subscription.instances, status, match_domain_attr=False)  # type:ignore

        try:
            model = cls(
                product=product,
                customer_id=subscription.customer_id,
                subscription_id=subscription.subscription_id,
                description=subscription.description,
                status=status,
                insync=subscription.insync,
                start_date=subscription.start_date,
                end_date=subscription.end_date,
                note=subscription.note,
                version=subscription.version,
                **fixed_inputs,
                **instances,
            )
            model.db_model = subscription
            return model
        except ValidationError:
            logger.exception(
                "Subscription is not correct in database", loaded_fixed_inputs=fixed_inputs, loaded_instances=instances
            )
            raise

    @classmethod
    def from_subscription(cls: type[S], subscription_id: UUID | UUIDstr) -> S:
        """Use a subscription_id to return required fields of an existing subscription."""
        from orchestrator.domain.context_cache import get_from_cache, store_in_cache

        if cached_model := get_from_cache(subscription_id):
            return cast(S, cached_model)

        if not (subscription := cls._get_subscription(subscription_id)):
            raise ValueError(f"Subscription with id: {subscription_id}, does not exist")
        product = cls._to_product_model(subscription.product)

        status = SubscriptionLifecycle(subscription.status)

        if not cls.__base_type__:
            # Import here to prevent cyclic imports
            from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY

            try:
                cls = SUBSCRIPTION_MODEL_REGISTRY[subscription.product.name]  # type:ignore
            except KeyError:
                raise ProductNotInRegistryError(
                    f"'{subscription.product.name}' is not found within the SUBSCRIPTION_MODEL_REGISTRY"
                )
            cls = lookup_specialized_type(cls, status)
        elif not issubclass(cls, lookup_specialized_type(cls, status)):
            raise ValueError(f"{cls} is not valid for lifecycle {status}")

        fixed_inputs = {fi.name: fi.value for fi in subscription.product.fixed_inputs}

        instances = cls._load_root_instances(subscription_id)

        try:
            model = cls(
                product=product,
                customer_id=subscription.customer_id,
                subscription_id=subscription.subscription_id,
                description=subscription.description,
                status=status,
                insync=subscription.insync,
                start_date=subscription.start_date,
                end_date=subscription.end_date,
                note=subscription.note,
                version=subscription.version,
                **fixed_inputs,
                **instances,
            )
            model.db_model = subscription

            store_in_cache(model)

            return model
        except ValidationError:
            logger.exception(
                "Subscription is not correct in database", loaded_fixed_inputs=fixed_inputs, loaded_instances=instances
            )
            raise

    def save(self) -> None:
        """Save the subscription to the database."""
        specialized_type = lookup_specialized_type(self.__class__, self.status)
        if specialized_type and not isinstance(self, specialized_type):
            raise ValueError(
                f"Lifecycle status {self.status.value} requires specialized type {specialized_type!r}, was: {type(self)!r}"
            )

        existing_sub = db.session.get(
            SubscriptionTable,
            self.subscription_id,
            options=[
                selectinload(SubscriptionTable.instances)
                .joinedload(SubscriptionInstanceTable.product_block)
                .selectinload(ProductBlockTable.resource_types),
                selectinload(SubscriptionTable.instances).selectinload(SubscriptionInstanceTable.values),
            ],
        )
        if not (sub := (existing_sub or self.db_model)):
            raise ValueError("Cannot save SubscriptionModel without a db_model")

        # Make sure we refresh the object and not use an already mapped object
        db.session.refresh(sub)

        self.db_model = sub
        sub.product_id = self.product.product_id
        sub.customer_id = self.customer_id
        sub.description = self.description
        sub.status = self.status.value
        sub.insync = self.insync
        sub.start_date = self.start_date
        sub.end_date = self.end_date
        sub.note = self.note

        db.session.add(sub)
        db.session.flush()  # Sends INSERT and returns subscription_id without committing transaction

        old_instances_dict = {instance.subscription_instance_id: instance for instance in sub.instances}

        saved_instances, depends_on_instances = self._save_instances(self.subscription_id, self.status)

        for instances in depends_on_instances.values():
            for instance in instances:
                if instance.subscription_id != self.subscription_id:
                    raise ValueError(
                        "Attempting to save a Foreign `Subscription Instance` directly below a subscription. "
                        "This is not allowed."
                    )
        sub.instances = saved_instances

        # Calculate what to remove
        instances_set = {instance.subscription_instance_id for instance in sub.instances}
        for instance_id in instances_set:
            old_instances_dict.pop(instance_id, None)

        # What's left should be removed
        for instance in old_instances_dict.values():
            db.session.delete(instance)

        db.session.flush()

    @property
    def db_model(self) -> SubscriptionTable | None:
        if not self._db_model:
            self._db_model = self._get_subscription(self.subscription_id)
        return self._db_model

    @db_model.setter
    def db_model(self, value: SubscriptionTable) -> None:
        self._db_model = value


def validate_base_model(
    name: str, cls: type[Any], base_model: type[BaseModel] = DomainModel, errors: list[str] | None = None
) -> None:
    """Validates that the given class is not Pydantic BaseModel or its direct subclass."""
    # Instantiate errors list if not provided and avoid mutating default
    if errors is None:
        errors = []
    # Return early when the node is not a class as there is nothing to be done
    if not isclass(cls):
        return
    # Validate each field in the ProductBlockModel's field dictionaries
    if issubclass(cls, ProductBlockModel) or issubclass(cls, SubscriptionModel):
        for name, clz in cls._product_block_fields_.items():
            validate_base_model(name, clz, ProductBlockModel, errors)
        for name, clz in cls._non_product_block_fields_.items():
            validate_base_model(name, clz, SubscriptionModel, errors)
    # Generate error if node is Pydantic BaseModel or direct subclass
    if issubclass(cls, BaseModel):
        err_msg: str = (
            f"If this field was intended to be a {base_model.__name__}, define {name}:{cls.__name__} with "
            f"{base_model.__name__} as its superclass instead. e.g., class {cls.__name__}({base_model.__name__}):"
        )
        if cls is BaseModel:
            errors.append(f"Field {name}: {cls.__name__} can not be {BaseModel.__name__}. " + err_msg)
        if len(cls.__mro__) > 1 and cls.__mro__[1] is BaseModel:
            errors.append(
                f"Field {name}: {cls.__name__} can not be a direct subclass of {BaseModel.__name__}. " + err_msg
            )
    # Format all errors as one per line and raise a TypeError when they exist
    if errors:
        raise TypeError("\n".join(errors))


class SubscriptionModelRegistry(dict[str, type[SubscriptionModel]]):
    """A registry for all subscription models."""

    def __setitem__(self, __key: str, __value: type[SubscriptionModel]) -> None:
        """Set value for key in while validating against Pydantic BaseModel."""
        validate_base_model(__key, __value)
        super().__setitem__(__key, __value)

    def update(
        self,
        m: Any = None,
        /,
        **kwargs: type[SubscriptionModel],
    ) -> None:
        """Update dictionary with mapping and/or kwargs using `__setitem__`."""
        if m:
            if isinstance(m, Mapping):
                for key, value in m.items():
                    self[key] = value
            elif isinstance(m, Iterable):
                for index, item in enumerate(m):
                    try:
                        key, value = item
                    except ValueError:
                        raise TypeError(f"dictionary update sequence element #{index} is not an iterable of length 2")
                    self[key] = value
        for key, value in kwargs.items():
            self[key] = value


def _validate_lifecycle_change_for_product_block(
    used_by: SubscriptionInstanceTable,
    product_block_model: ProductBlockModel,
    status: SubscriptionLifecycle,
    description: str,
) -> None:
    """Validate if a lifecycle change for a single product model is possible."""

    logger.debug(
        "Checking the parent relations",
        parent_status=used_by.subscription.status,
        parent_description=used_by.subscription.description,
        self_status=status,
        self_description=description,
    )
    if (
        used_by.subscription != product_block_model.subscription
        and used_by.subscription.status not in SAFE_USED_BY_TRANSITIONS_FOR_STATUS[status]
    ):
        raise ValueError(
            f"Unsafe status change of Subscription with depending subscriptions: "
            f"{list(map(lambda instance: description, product_block_model.in_use_by))}"
        )


def validate_lifecycle_change(
    other: "SubscriptionModel",
    status: SubscriptionLifecycle,
) -> None:
    """Validate if a lifecycle change for a subscription model is possible.

    It will traverse all product blocks and check the `in_use_by` status to ensure that the lifecycle change
    is allowed.

    Note: A `ValueError` will be raised when a unsafe status change is found
    """
    for product_block_field, product_block_field_type in other._product_block_fields_.items():
        product_block_models = getattr(other, product_block_field)
        if is_list_type(product_block_field_type):
            for product_block_model in product_block_models:
                used_by_generator = (used_by for used_by in product_block_model.in_use_by if used_by)
                for used_by in used_by_generator:
                    _validate_lifecycle_change_for_product_block(
                        used_by, product_block_model, status, other.description
                    )
        elif (
            is_optional_type(product_block_field_type) or is_union_type(product_block_field_type)
        ) and product_block_models is None:
            pass
        else:
            used_by_generator = (used_by for used_by in product_block_models.in_use_by if used_by)
            for used_by in used_by_generator:
                _validate_lifecycle_change_for_product_block(used_by, product_block_models, status, other.description)

    logger.info(
        "Lifecycle validation check ok",
        subscription_id=other.subscription_id,
        subscription_description=other.description,
        status=status,
    )



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from itertools import count
from typing import Any, ClassVar, Literal

from pydantic import BaseModel, ConfigDict, Field, model_validator
from sqlalchemy import BinaryExpression, and_, cast, exists, literal, or_, select
from sqlalchemy.dialects.postgresql import BOOLEAN
from sqlalchemy.sql.elements import ColumnElement
from sqlalchemy_utils.types.ltree import Ltree

from orchestrator.db.models import AiSearchIndex
from orchestrator.search.core.types import BooleanOperator, FieldType, FilterOp, SQLAColumn, UIType

from .date_filters import DateFilter
from .ltree_filters import LtreeFilter
from .numeric_filter import NumericFilter


class EqualityFilter(BaseModel):
    op: Literal[FilterOp.EQ, FilterOp.NEQ]
    value: Any

    def to_expression(self, column: SQLAColumn, path: str) -> BinaryExpression[bool] | ColumnElement[bool]:
        if isinstance(self.value, bool):
            colb = cast(column, BOOLEAN)
            return colb.is_(self.value) if self.op == FilterOp.EQ else ~colb.is_(self.value)
        sv = str(self.value)
        return (column == sv) if self.op == FilterOp.EQ else (column != sv)


class StringFilter(BaseModel):
    op: Literal[FilterOp.LIKE]
    value: str

    def to_expression(self, column: SQLAColumn, path: str) -> ColumnElement[bool]:
        return column.like(self.value)

    @model_validator(mode="after")
    def validate_like_pattern(self) -> StringFilter:
        """If the operation is 'like', the value must contain a wildcard."""
        if self.op == FilterOp.LIKE:
            if "%" not in self.value and "_" not in self.value:
                raise ValueError("The value for a 'like' operation must contain a wildcard character ('%' or '_').")
        return self


# Order matters! Ambiguous ops (like 'eq') are resolved by first matching filter
FilterCondition = (
    DateFilter  # DATETIME
    | NumericFilter  # INT/FLOAT
    | StringFilter  # STRING TODO: convert to hybrid search?
    | LtreeFilter  # Path
    | EqualityFilter  # BOOLEAN/UUID/BLOCK/RESOURCE_TYPE - most generic, try last
)


class PathFilter(BaseModel):

    path: str = Field(description="The ltree path of the field to filter on, e.g., 'subscription.customer_id'.")
    condition: FilterCondition = Field(description="The filter condition to apply.")

    value_kind: UIType

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {"path": "subscription.status", "condition": {"op": "eq", "value": "active"}, "value_kind": "string"},
                {
                    "path": "subscription.customer_id",
                    "condition": {"op": "neq", "value": "acme"},
                    "value_kind": "string",
                },
                {
                    "path": "subscription.start_date",
                    "condition": {"op": "gt", "value": "2025-01-01"},
                    "value_kind": "datetime",
                },
                {
                    "path": "subscription.end_date",
                    "condition": {
                        "op": "between",
                        "value": {"start": "2025-06-01", "end": "2025-07-01"},
                    },
                    "value_kind": "datetime",
                },
                {
                    "path": "subscription",
                    "condition": {"op": "has_component", "value": "node"},
                    "value_kind": "component",
                },
            ]
        }
    )

    @model_validator(mode="before")
    @classmethod
    def _transfer_path_to_value_if_needed(cls, data: Any) -> Any:
        """Transform for path-only filters.

        If `op` is `has_component`, `not_has_component`, or `ends_with` and no `value` is
        provided in the `condition`, this validator will automatically use the `path`
        field as the `value` and set the `path` to a wildcard '*' for the query.
        """
        if isinstance(data, dict):
            path = data.get("path")
            condition = data.get("condition")

            if path and isinstance(condition, dict):
                op = condition.get("op")
                value = condition.get("value")

                path_only_ops = [FilterOp.HAS_COMPONENT, FilterOp.NOT_HAS_COMPONENT, FilterOp.ENDS_WITH]

                if op in path_only_ops and value is None:
                    condition["value"] = path
                    data["path"] = "*"
        return data

    def to_expression(self, value_column: SQLAColumn, value_type_column: SQLAColumn) -> ColumnElement[bool]:
        """Convert the path filter into a SQLAlchemy expression with type safety.

        This method creates a type guard to ensure we only match compatible field types,
        then delegates to the specific filter condition.

        Args:
            value_column (ColumnElement): The SQLAlchemy column element representing the value to be filtered.
            value_type_column (ColumnElement): The SQLAlchemy column element representing the field type.

        Returns:
            ColumnElement[bool]: A SQLAlchemy boolean expression that can be used in a ``WHERE`` clause.
        """

        # Type guard - only match compatible field types
        allowed_field_types = [ft.value for ft in FieldType if UIType.from_field_type(ft) == self.value_kind]
        type_guard = value_type_column.in_(allowed_field_types) if allowed_field_types else literal(True)

        return and_(type_guard, self.condition.to_expression(value_column, self.path))


class FilterTree(BaseModel):
    op: BooleanOperator = Field(
        description="Operator for grouping conditions in uppercase.", default=BooleanOperator.AND
    )

    children: list[FilterTree | PathFilter] = Field(min_length=1, description="Path filters or nested groups.")

    MAX_DEPTH: ClassVar[int] = 5

    model_config = ConfigDict(
        json_schema_extra={
            "description": (
                "Boolean filter tree. Operators must be UPPERCASE: AND / OR.\n"
                "Node shapes:\n"
                "  • Group: {'op':'AND'|'OR', 'children': [<PathFilter|FilterTree>, ...]}\n"
                "  • Leaf (PathFilter): {'path':'<ltree>', 'condition': {...}}\n"
                "Rules:\n"
                "  • Do NOT put 'op' or 'children' inside a leaf 'condition'.\n"
                f"  • Max depth = {MAX_DEPTH}.\n"
            ),
            "examples": [
                {
                    "description": "Simple filters",
                    "op": "AND",
                    "children": [
                        {"path": "subscription.status", "condition": {"op": "eq", "value": "active"}},
                        {"path": "subscription.start_date", "condition": {"op": "gt", "value": "2021-01-01"}},
                    ],
                },
                {
                    "description": "Complex filters with OR group",
                    "op": "AND",
                    "children": [
                        {"path": "subscription.start_date", "condition": {"op": "gte", "value": "2024-01-01"}},
                        {
                            "op": "OR",
                            "children": [
                                {"path": "subscription.product.name", "condition": {"op": "like", "value": "%fiber%"}},
                                {"path": "subscription.customer_id", "condition": {"op": "eq", "value": "Surf"}},
                            ],
                        },
                    ],
                },
            ],
        }
    )

    @model_validator(mode="after")
    def _validate_depth(self) -> FilterTree:
        def depth(node: "FilterTree | PathFilter") -> int:
            return 1 + max(depth(c) for c in node.children) if isinstance(node, FilterTree) else 1

        if depth(self) > self.MAX_DEPTH:
            raise ValueError(f"FilterTree nesting exceeds MAX_DEPTH={self.MAX_DEPTH}")
        return self

    @classmethod
    def from_flat_and(cls, filters: list[PathFilter]) -> FilterTree | None:
        """Wrap a flat list of PathFilter into an AND group (or None)."""
        return None if not filters else cls(op=BooleanOperator.AND, children=list(filters))

    def get_all_paths(self) -> set[str]:
        """Collects all unique paths from the PathFilter leaves in the tree."""
        return {leaf.path for leaf in self.get_all_leaves()}

    def get_all_leaves(self) -> list[PathFilter]:
        """Collect all PathFilter leaves in the tree."""
        leaves: list[PathFilter] = []
        for child in self.children:
            if isinstance(child, PathFilter):
                leaves.append(child)
            else:
                leaves.extend(child.get_all_leaves())
        return leaves

    def to_expression(
        self,
        entity_id_col: SQLAColumn,
        *,
        entity_type_value: str | None = None,
    ) -> ColumnElement[bool]:
        """Compile this tree into a SQLAlchemy boolean expression.

        Args:
            entity_id_col (SQLAColumn): Column in the outer query representing the entity ID.
            entity_type_value (str, optional): If provided, each subquery is additionally constrained to this entity type.

        Returns:
            ColumnElement[bool]: A SQLAlchemy expression suitable for use in a WHERE clause.
        """

        alias_idx = count(1)

        def leaf_exists(pf: PathFilter) -> ColumnElement[bool]:
            from sqlalchemy.orm import aliased

            alias = aliased(AiSearchIndex, name=f"flt_{next(alias_idx)}")

            correlates = [alias.entity_id == entity_id_col]
            if entity_type_value is not None:
                correlates.append(alias.entity_type == entity_type_value)

            if isinstance(pf.condition, LtreeFilter):
                # row-level predicate is always positive
                positive = pf.condition.to_expression(alias.path, pf.path)
                subq = select(1).select_from(alias).where(and_(*correlates, positive))
                if pf.condition.op == FilterOp.NOT_HAS_COMPONENT:
                    return ~exists(subq)  # NOT at the entity level
                return exists(subq)

            # value leaf: path predicate + typed value compare
            if "." not in pf.path:
                path_pred = LtreeFilter(op=FilterOp.ENDS_WITH, value=pf.path).to_expression(alias.path, "")
            else:
                path_pred = alias.path == Ltree(pf.path)

            value_pred = pf.to_expression(alias.value, alias.value_type)
            subq = select(1).select_from(alias).where(and_(*correlates, path_pred, value_pred))
            return exists(subq)

        def compile_node(node: FilterTree | PathFilter) -> ColumnElement[bool]:
            if isinstance(node, FilterTree):
                compiled = [compile_node(ch) for ch in node.children]
                return and_(*compiled) if node.op == BooleanOperator.AND else or_(*compiled)
            return leaf_exists(node)

        return compile_node(self)



from sqlalchemy.orm import Load

from orchestrator.db.database import BaseModel as DbBaseModel
from orchestrator.db.loaders import (
    get_query_loaders_for_model_paths,
)
from orchestrator.graphql.types import OrchestratorInfo
from orchestrator.graphql.utils.get_selected_paths import get_selected_paths


def get_query_loaders_for_gql_fields(root_model: type[DbBaseModel], info: OrchestratorInfo) -> list[Load]:
    """Get sqlalchemy query loaders for the given GraphQL query.

    Based on the GraphQL query's selected fields, returns the required DB loaders to use
    in SQLALchemy's `.options()` for efficiently quering (nested) relationships.
    """
    model_paths = [path.removeprefix("page.") for path in get_selected_paths(info)]

    return get_query_loaders_for_model_paths(root_model, model_paths)



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from decimal import Decimal

import structlog
from sqlalchemy import BindParameter, Numeric, Select, literal

from orchestrator.search.core.types import FieldType, SearchMetadata
from orchestrator.search.schemas.parameters import BaseSearchParameters

from ..pagination import PaginationParams

logger = structlog.get_logger(__name__)


class Retriever(ABC):
    """Abstract base class for applying a ranking strategy to a search query."""

    SCORE_PRECISION = 12
    SCORE_NUMERIC_TYPE = Numeric(38, 12)
    HIGHLIGHT_TEXT_LABEL = "highlight_text"
    HIGHLIGHT_PATH_LABEL = "highlight_path"
    SCORE_LABEL = "score"
    SEARCHABLE_FIELD_TYPES = [
        FieldType.STRING.value,
        FieldType.UUID.value,
        FieldType.BLOCK.value,
        FieldType.RESOURCE_TYPE.value,
    ]

    @classmethod
    async def from_params(
        cls,
        params: BaseSearchParameters,
        pagination_params: PaginationParams,
    ) -> "Retriever":
        """Create the appropriate retriever instance from search parameters.

        Args:
            params (BaseSearchParameters): Search parameters including vector queries, fuzzy terms, and filters.
            pagination_params (PaginationParams): Pagination parameters for cursor-based paging.

        Returns:
            Retriever: A concrete retriever instance (semantic, fuzzy, hybrid, or structured).
        """

        from .fuzzy import FuzzyRetriever
        from .hybrid import RrfHybridRetriever
        from .semantic import SemanticRetriever
        from .structured import StructuredRetriever

        fuzzy_term = params.fuzzy_term
        q_vec = await cls._get_query_vector(params.vector_query, pagination_params.q_vec_override)

        # If semantic search was attempted but failed, fall back to fuzzy with the full query
        fallback_fuzzy_term = fuzzy_term
        if q_vec is None and params.vector_query is not None and params.query is not None:
            fallback_fuzzy_term = params.query

        if q_vec is not None and fallback_fuzzy_term is not None:
            return RrfHybridRetriever(q_vec, fallback_fuzzy_term, pagination_params)
        if q_vec is not None:
            return SemanticRetriever(q_vec, pagination_params)
        if fallback_fuzzy_term is not None:
            return FuzzyRetriever(fallback_fuzzy_term, pagination_params)

        return StructuredRetriever(pagination_params)

    @classmethod
    async def _get_query_vector(
        cls, vector_query: str | None, q_vec_override: list[float] | None
    ) -> list[float] | None:
        """Get query vector either from override or by generating from text."""
        if q_vec_override:
            return q_vec_override

        if not vector_query:
            return None

        from orchestrator.search.core.embedding import QueryEmbedder

        q_vec = await QueryEmbedder.generate_for_text_async(vector_query)
        if not q_vec:
            logger.warning("Embedding generation failed; using non-semantic retriever")
            return None

        return q_vec

    @abstractmethod
    def apply(self, candidate_query: Select) -> Select:
        """Apply the ranking logic to the given candidate query.

        Args:
            candidate_query (Select): A SQLAlchemy `Select` statement returning candidate entity IDs.

        Returns:
            Select: A new `Select` statement with ranking expressions applied.
        """
        ...

    def _quantize_score_for_pagination(self, score_value: float) -> BindParameter[Decimal]:
        """Convert score value to properly quantized Decimal parameter for pagination."""
        pas_dec = Decimal(str(score_value)).quantize(Decimal("0.000000000001"))
        return literal(pas_dec, type_=self.SCORE_NUMERIC_TYPE)

    @property
    @abstractmethod
    def metadata(self) -> SearchMetadata:
        """Return metadata describing this search strategy."""
        ...



# Copyright 2022-2025 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections.abc import Iterator

from strawberry.extensions import SchemaExtension

from orchestrator.domain.context_cache import cache_subscription_models


class ModelCacheExtension(SchemaExtension):
    """Wraps the GraphQL operation in a cache_subscription_models context.

    For more background, please refer to the documentation of the contextmanager.
    """

    def on_operation(self, *args, **kwargs) -> Iterator[None]:  # type: ignore

        with cache_subscription_models():
            yield



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import structlog
from fastapi import FastAPI, HTTPException
from pydantic_ai.ag_ui import StateDeps
from pydantic_ai.agent import Agent
from pydantic_ai.models.openai import OpenAIModel
from pydantic_ai.settings import ModelSettings
from pydantic_ai.toolsets import FunctionToolset
from starlette.types import ASGIApp

from orchestrator.search.agent.prompts import get_base_instructions, get_dynamic_instructions
from orchestrator.search.agent.state import SearchState
from orchestrator.search.agent.tools import search_toolset

logger = structlog.get_logger(__name__)


def _disabled_agent_app(reason: str) -> FastAPI:
    app = FastAPI(title="Agent disabled")

    @app.api_route("/{path:path}", methods=["GET", "POST", "PUT", "PATCH", "DELETE", "OPTIONS", "HEAD"])
    async def _disabled(path: str) -> None:
        raise HTTPException(status_code=503, detail=f"Agent disabled: {reason}")

    return app


def build_agent_app(model: str | OpenAIModel, toolsets: list[FunctionToolset[Any]] | None = None) -> ASGIApp:
    try:
        toolsets = toolsets + [search_toolset] if toolsets else [search_toolset]

        agent = Agent(
            model=model,
            deps_type=StateDeps[SearchState],
            model_settings=ModelSettings(
                parallel_tool_calls=False,
            ),  # https://github.com/pydantic/pydantic-ai/issues/562
            toolsets=toolsets,
        )
        agent.instructions(get_base_instructions)
        agent.instructions(get_dynamic_instructions)

        return agent.to_ag_ui(deps=StateDeps(SearchState()))
    except Exception as e:
        logger.error("Agent init failed; serving disabled stub.", error=str(e))
        return _disabled_agent_app(str(e))



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import secrets
import string
from pathlib import Path
from typing import Literal

from pydantic import Field, NonNegativeInt, PostgresDsn, RedisDsn
from pydantic_settings import BaseSettings

from oauth2_lib.settings import oauth2lib_settings
from orchestrator.services.settings_env_variables import expose_settings
from orchestrator.utils.expose_settings import SecretStr as OrchSecretStr
from pydantic_forms.types import strEnum


class ExecutorType(strEnum):
    WORKER = "celery"
    THREADPOOL = "threadpool"


class AppSettings(BaseSettings):
    TESTING: bool = True
    SESSION_SECRET: OrchSecretStr = "".join(secrets.choice(string.ascii_letters) for i in range(16))  # type: ignore
    CORS_ORIGINS: str = "*"
    CORS_ALLOW_METHODS: list[str] = ["GET", "PUT", "PATCH", "POST", "DELETE", "OPTIONS", "HEAD"]
    CORS_ALLOW_HEADERS: list[str] = ["If-None-Match", "Authorization", "If-Match", "Content-Type"]
    CORS_EXPOSE_HEADERS: list[str] = [
        "Cache-Control",
        "Content-Language",
        "Content-Length",
        "Content-Type",
        "Expires",
        "Last-Modified",
        "Pragma",
        "Content-Range",
        "ETag",
    ]
    ENVIRONMENT: str = "local"
    EXECUTOR: str = ExecutorType.THREADPOOL
    WORKFLOWS_SWAGGER_HOST: str = "localhost"
    WORKFLOWS_GUI_URI: str = "http://localhost:3000"
    DATABASE_URI: PostgresDsn = "postgresql://nwa:nwa@localhost/orchestrator-core"  # type: ignore
    MAX_WORKERS: int = 5
    MAIL_SERVER: str = "localhost"
    MAIL_PORT: int = 25
    MAIL_STARTTLS: bool = False
    CACHE_URI: RedisDsn = "redis://localhost:6379/0"  # type: ignore
    CACHE_HMAC_SECRET: OrchSecretStr | None = None  # HMAC signing key, used when pickling results in the cache
    REDIS_RETRY_COUNT: NonNegativeInt = Field(
        2, description="Number of retries for redis connection errors/timeouts, 0 to disable"
    )  # More info: https://redis-py.readthedocs.io/en/stable/retry.html
    ENABLE_DISTLOCK_MANAGER: bool = True
    DISTLOCK_BACKEND: str = "memory"
    CC_NOC: int = 0
    SERVICE_NAME: str = "orchestrator-core"
    LOGGING_HOST: str = "localhost"
    LOG_LEVEL: str = "DEBUG"
    SLACK_ENGINE_SETTINGS_HOOK_ENABLED: bool = False
    SLACK_ENGINE_SETTINGS_HOOK_URL: str = ""
    TRACING_ENABLED: bool = False
    TRACE_HOST: str = "http://localhost:4317"
    TRANSLATIONS_DIR: Path | None = None
    WEBSOCKET_BROADCASTER_URL: OrchSecretStr = "memory://"  # type: ignore
    ENABLE_WEBSOCKETS: bool = True
    DISABLE_INSYNC_CHECK: bool = False
    DEFAULT_PRODUCT_WORKFLOWS: list[str] = ["modify_note"]
    SKIP_MODEL_FOR_MIGRATION_DB_DIFF: list[str] = []
    SERVE_GRAPHQL_UI: bool = True
    FEDERATION_ENABLED: bool = False
    DEFAULT_CUSTOMER_FULLNAME: str = "Default::Orchestrator-Core Customer"
    DEFAULT_CUSTOMER_SHORTCODE: str = "default-cust"
    DEFAULT_CUSTOMER_IDENTIFIER: str = "59289a57-70fb-4ff5-9c93-10fe67b12434"
    TASK_LOG_RETENTION_DAYS: int = 3
    ENABLE_GRAPHQL_DEPRECATION_CHECKER: bool = True
    ENABLE_GRAPHQL_PROFILING_EXTENSION: bool = False
    ENABLE_GRAPHQL_STATS_EXTENSION: bool = False
    ENABLE_PROMETHEUS_METRICS_ENDPOINT: bool = False
    VALIDATE_OUT_OF_SYNC_SUBSCRIPTIONS: bool = False
    FILTER_BY_MODE: Literal["partial", "exact"] = "exact"
    EXPOSE_SETTINGS: bool = False
    EXPOSE_OAUTH_SETTINGS: bool = False


app_settings = AppSettings()

# Set oauth2lib_settings variables to the same (default) value of settings
oauth2lib_settings.SERVICE_NAME = app_settings.SERVICE_NAME
oauth2lib_settings.ENVIRONMENT = app_settings.ENVIRONMENT

if app_settings.EXPOSE_SETTINGS:
    expose_settings("app_settings", app_settings)  # type: ignore
if app_settings.EXPOSE_OAUTH_SETTINGS:
    expose_settings("oauth2lib_settings", oauth2lib_settings)  # type: ignore



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import ClassVar

from pydantic_forms.core import DisplayOnlyFieldType, generate_form, post_form
from pydantic_forms.core import FormPage as PydanticFormsFormPage
from pydantic_forms.types import JSON, InputForm, StateInputFormGenerator

__all__ = [
    "DisplayOnlyFieldType",
    "FormPage",
    "SubmitFormPage",
    "InputForm",
    "JSON",
    "StateInputFormGenerator",
    "generate_form",
    "post_form",
]


class FormPage(PydanticFormsFormPage):
    meta__: ClassVar[JSON] = {"hasNext": True}


class SubmitFormPage(FormPage):
    meta__: ClassVar[JSON] = {"hasNext": False}



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
from collections.abc import Generator, Iterable
from importlib import import_module
from os import listdir, path
from pathlib import Path
from typing import Any

import structlog
from more_itertools import first, one

from orchestrator.cli.generator.generator.enums import to_dict
from orchestrator.cli.generator.generator.settings import product_generator_settings as settings
from orchestrator.domain.base import ProductBlockModel
from orchestrator.utils.helpers import camel_to_snake, snake_to_camel

logger = structlog.getLogger(__name__)


def get_workflow(config: dict, workflow_name: str) -> dict:
    workflows = (workflow for workflow in config.get("workflows", []) if workflow["name"] == workflow_name)
    return first(workflows, {})


def get_variable(config: dict) -> str:
    return config.get("variable", camel_to_snake(config["name"]))


def get_product_block_variable(product_block: dict) -> str:
    return get_variable(product_block)


def get_product_file_name(config: dict) -> str:
    return get_variable(config)


def get_product_block_file_name(product_block: dict) -> str:
    return get_product_block_variable(product_block)


def get_existing_product_blocks() -> dict[str, Any]:
    """Inspect the python code for existing product blocks."""

    def yield_blocks() -> Generator:
        def is_product_block(attribute: Any) -> bool:
            return issubclass(attribute, ProductBlockModel)

        if not path.exists(get_product_blocks_folder()):
            logger.warning("Product block path does not exist", product_blocks_path=get_product_blocks_folder())
            return

        for pb_file in listdir(get_product_blocks_folder()):
            name = pb_file.removesuffix(".py")
            module_name = f"{get_product_blocks_module()}.{name}"

            module = import_module(module_name)

            classes = [obj for _, obj in inspect.getmembers(module, inspect.isclass) if obj.__module__ == module_name]

            yield from ((klass.__name__, module_name) for klass in classes if is_product_block(klass))

    return dict(yield_blocks())


def get_product_block_depends_on(
    product_blocks: list[dict], include_existing_blocks: bool = False
) -> dict[str, set[str]]:
    _product_block_types = {block["type"] for block in product_blocks}

    def base_type(block_name: str) -> str:
        block_type, _lifecycle = block_name.rsplit("Block", maxsplit=1)
        return block_type

    if include_existing_blocks:
        existing_blocks = {base_type(block) for block in get_existing_product_blocks()}
        _product_block_types.update(existing_blocks)

    def dependencies(product_block: dict) -> Iterable[str]:
        """Find all product blocks which this product block depends on."""
        for field in product_block.get("fields", []):
            field_type = field.get("list_type", field["type"])
            if field_type in _product_block_types:
                yield field_type

    return {product_block["type"]: set(dependencies(product_block)) for product_block in product_blocks}


def find_root_product_block(product_blocks: list[dict]) -> str | None:
    block_dependencies = get_product_block_depends_on(product_blocks)

    blocks_in_use = set().union(*block_dependencies.values())
    root_blocks = block_dependencies.keys() - blocks_in_use
    return one(
        root_blocks,
        too_short=ValueError(
            "There should be exactly 1 root product block, found none. Please ensure there are no cyclic relations"
        ),
        too_long=ValueError(f"There should be exactly 1 root product block, found multiple: {root_blocks}"),
    )


def root_product_block(config: dict) -> dict:
    product_blocks = config.get("product_blocks", [])
    root_block_name = find_root_product_block(config.get("product_blocks", []))
    return one(block for block in product_blocks if block["type"] == root_block_name)


def sort_product_blocks_by_dependencies(product_blocks: list[dict]) -> list[dict]:
    """Perform a 'Topological Sort' on the list of product blocks.

    This ensures that a product's blocks are created bottom-up and that there is no cycle.
    """
    block_dependencies = get_product_block_depends_on(product_blocks)

    block_order: dict[str, int] = {}
    order = 0
    while block_dependencies:
        cycle = True

        for block, depends_on_blocks in list(block_dependencies.items()):
            if depends_on_blocks - block_order.keys():
                # Not all dependent blocks are resolved yet
                continue
            cycle = False
            block_order[block] = order
            order += 1
            del block_dependencies[block]

        if cycle:
            raise ValueError(f"Cycle detected in product blocks: {block_dependencies}")

    return sorted(product_blocks, key=lambda block: block_order[block["type"]])


def set_resource_types(product_blocks: list[dict], block_dependencies: dict[str, set[str]]) -> list[dict]:
    """Returns product blocks enriched with a list 'resource_types'.

    Args:
        product_blocks: product blocks to enrich
        block_dependencies: mapping of product blocks to dependent blocks
    """

    def resource_type_fields(product_block: dict) -> Iterable[dict]:
        for field in product_block["fields"]:
            field_type = field.get("list_type", field["type"])
            if field_type not in block_dependencies[product_block["type"]]:
                yield field

    return [(block | {"resource_types": list(resource_type_fields(block))}) for block in product_blocks]


def insert_into_imports(content: list[str], new_import: str) -> list[str]:
    # Note: we may consider using a real Python parser here someday, but for now this is ok and formatting
    # gets done by isort and black.
    def produce() -> Generator:
        not_inserted_yet = True
        for line in content:
            if line.startswith("from ") and not_inserted_yet:
                yield new_import
                not_inserted_yet = False
            yield line

    return list(produce())


def path_to_module(path: Path) -> str:
    return str(path).replace("/", ".")


def get_workflows_folder() -> Path:
    return settings.FOLDER_PREFIX / settings.WORKFLOWS_PATH


def get_product_blocks_folder() -> Path:
    return settings.FOLDER_PREFIX / settings.PRODUCT_BLOCKS_PATH


def get_product_types_folder() -> Path:
    return settings.FOLDER_PREFIX / settings.PRODUCT_TYPES_PATH


def get_workflows_module() -> str:
    return path_to_module(get_workflows_folder())


def get_product_blocks_module() -> str:
    return path_to_module(get_product_blocks_folder())


def get_product_types_module() -> str:
    return path_to_module(get_product_types_folder())


def get_product_path(config: dict) -> Path:
    file_name = get_product_file_name(config)
    return get_product_types_folder() / Path(file_name).with_suffix(".py")


def get_product_import(product: dict, lifecycle: str = "") -> str:
    return f'from {get_product_types_module()}.{product["variable"]} import {product["type"]}{lifecycle}\n'


def create_dunder_init_files(path: Path) -> None:
    folder = Path("")
    for part in path.parts:
        if (folder := folder / part).is_dir():
            if not (dunder_init_file := folder / Path("__init__.py")).exists():
                logger.info("creating missing dunder init", path=str(dunder_init_file))
                open(dunder_init_file, "x").close()


def is_constrained_int(field: dict) -> bool:
    return "min_value" in field or "max_value" in field


def get_constrained_ints(fields: list[dict]) -> list[dict]:
    return [field for field in fields if is_constrained_int(field)]


def merge_fields(fields: list[dict], int_enums: list[dict], str_enums: list[dict]) -> list[dict]:
    return list((to_dict(fields) | to_dict(int_enums) | to_dict(str_enums)).values())


def is_name_spaced_field_type(field: dict) -> bool:
    return "." in field["type"]


def name_space_get_type(name_spaced_type: str) -> str:
    return name_spaced_type.split(".")[-1]


def process_fields(fields: list[dict]) -> list[dict]:
    def to_type(field: dict) -> dict:
        if is_constrained_int(field):
            return field | {"type": snake_to_camel(field["name"])}

        if is_name_spaced_field_type(field):
            return field | {"type": name_space_get_type(field["type"])}

        return field

    return [to_type(field) for field in fields]


def get_all_fields(product_block: dict) -> list[dict]:
    return process_fields(product_block["fields"])


def get_input_fields(product_block: dict) -> list[dict]:
    def supported_input_type(field: dict) -> bool:
        return field["type"] in ("int", "str", "bool", "enum")

    return process_fields([field for field in product_block["fields"] if supported_input_type(field)])


def get_name_spaced_types_to_import(fields: list) -> list[tuple]:
    # NOTE: we could make this smarter by grouping imports from the namespace, but isort will handle this for us
    def name_space_split(field: dict) -> tuple[str, str]:
        *namespace, type = field["type"].split(".")
        return ".".join(namespace), type

    return [name_space_split(field) for field in fields if is_name_spaced_field_type(field)]
